from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable, Type, Set

# from .top import Industrial
# from omnidata.parameters.abstract import AbstractParameterized

from omnibelt.crafts import ProcessedCrafty, IndividualCrafty, AbstractCraft, AbstractSkill, AbstractCrafty

from ..structure import spaces
from ..tools.abstract import AbstractTool, AbstractContext, AbstractSpaced, AbstractChangableSpace, Gizmoed
from ..tools.errors import ToolFailedError
from ..tools.kits import SpaceKit, ElasticCrafty
from ..tools.context import DynamicContext
from ..tools.kits import CraftyKit
from ..tools.crafts import SpaceCraft
from ..tools import Industrial, Spatial

from .abstract import AbstractModular, AbstractBuilder, \
	AbstractSubmodule, AbstractArgumentBuilder, AbstractParameterized
from .parameterized import ParameterizedBase
from .building import BuilderBase


class AbstractSpecced(AbstractSpaced):
	# @property
	# def my_blueprint(self):
	# 	raise NotImplementedError


	# def as_spec(self) -> 'AbstractSpec':
	# 	raise NotImplementedError


	# def _missing_spaces(self) -> Iterator[str]:
	# 	yield from ()

	pass



class AbstractSpec(AbstractContext, AbstractChangableSpace):
	def sub(self, submodule) -> 'AbstractSpec':
		raise NotImplementedError


	def for_builder(self):
		'''Returns the spec for a builder (usually the builder is created automatically for a submodule)'''
		raise NotImplementedError



class SpecBase(DynamicContext, AbstractSpec):
	def __init__(self, *, spaces=None, **kwargs):
		super().__init__(**kwargs)
		# self._owner = owner
		self._spaces = {}
		if spaces is not None:
			self._spaces.update(spaces)


	def include(self, *sources: AbstractTool):
		for source in sources:
			self._integrate_source(source)
		return super().include(*sources)


	def _integrate_source(self, source: AbstractTool):
		if isinstance(source, AbstractSpecced):
			for gizmo in source.gizmos():
				try:
					self.space_of(gizmo)
				except ToolFailedError:
					self.change_space_of(gizmo, source.space_of(gizmo))


	@property
	def size(self):
		return 1
	@property
	def indices(self):
		return [0]


	def sub(self, submodule) -> 'AbstractSpec':
		return self


	def adapt(self, overrides):
		new = self.__class__(spaces=overrides)
		new.include(self)
		return new


	def for_builder(self):
		return self


	def copy(self, **kwargs):
		return self.__class__(spaces=self._spaces, **kwargs)


	def change_space_of(self, gizmo: str, space: spaces.Dim):
		self._spaces[gizmo] = space


	def gizmos(self) -> Iterator[str]:
		yield from self._spaces.keys()


	def has_gizmo(self, gizmo: str) -> bool:
		return gizmo in self._spaces


	def space_of(self, gizmo: str) -> spaces.Dim:
		if gizmo in self._spaces:
			return self._spaces[gizmo]
		return super().space_of(gizmo)


	# def update_with(self, other: 'AbstractSpecced'):
	# 	for gizmo in other.gizmos():
	# 		try:
	# 			space = other.space_of(gizmo)
	# 		except ToolFailedError:
	# 			continue
	# 		else:
	# 			try:
	# 				prev = self.space_of(gizmo)
	# 			except ToolFailedError:
	# 				self.change_space_of(gizmo, space)
	# 			else:
	# 				if prev is None:
	# 					self.change_space_of(gizmo, space)
	#
	# 	return self


# one of the deepest classes in the MRO (before hparam extraction and craft processing)

class PlannedBase(AbstractSpecced):
	_Spec = None
	def __init__(self, *args, blueprint=None, **kwargs):
		if blueprint is None:
			blueprint = self._Spec() # TODO: maybe include `self`
		super().__init__(*args, **kwargs)
		self._my_blueprint = blueprint


	@property
	def my_blueprint(self):
		return self._my_blueprint



class PlannedModules(PlannedBase, AbstractModular):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._initialize_submodules()


	def _initialize_submodules(self):
		for name, sub in self.named_submodules(hidden=True):
			sub.init(self)


	def _validate_with_spec(self, spec):
		pass


	def _validate_submodules(self, spec=None):
		if spec is None:
			spec = self.my_blueprint

		# explicitly requested spaces are inferred from spec and set to self (potentially using the subs)
		# add self to spec (implicitly adds all spaces defined in self to spec) (also means space wise self is ready)
		self._validate_with_spec(spec)
		spec.include(self)

		# if sub is missing or a builder, this will instantiate it (using the sub-spec)
		for name, sub in self.named_submodules(hidden=True):
			sub.validate(self)



class PlannedSpatial(Spatial, PlannedModules): # could still be a builder
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._validate_submodules()


	def _validate_with_spec(self, spec):
		for gizmo in self._missing_spaces():
			space = spec.space_of(gizmo)
			if space is not None:
				self.change_space_of(gizmo, space)


	def _missing_spaces(self):
		for gizmo, opts in self._spaces.items():
			if not len(opts):
				yield gizmo
			elif opts[0].is_missing(gizmo):
				yield gizmo



class PlannedIndustrial(Industrial, PlannedSpatial):
	pass



class Specced(PlannedIndustrial):
	pass



class ArchitectBase(PlannedSpatial, AbstractArgumentBuilder):
	def _build_kwargs(self, product, *args, **kwargs):
		kwargs = super()._build_kwargs(product, *args, **kwargs)
		if issubclass(product, Industrial) and 'application' not in kwargs and self._application is not None:
			kwargs['application'] = self._application
		if issubclass(product, Specced) and 'blueprint' not in kwargs and self.my_blueprint is not None:
			kwargs['blueprint'] = self.my_blueprint
		return kwargs


	def _validate_with_spec(self, spec):
		for gizmo in self._missing_spaces():
			try:
				space = spec.space_of(gizmo)
			except ToolFailedError:
				pass
			else:
				if space is not None:
					self.change_space_of(gizmo, space)


	class _Plan:
		arch: 'ArchitectBase'

		def __init__(self, arch, blueprint):
			self.arch = arch
			self.spec = blueprint
			self.fixes = None


		def __enter__(self):
			fixes = []
			for gizmo in self.arch._missing_spaces():
				space = self.spec.space_of(gizmo)
				if space is not None:
					self.arch.change_space_of(gizmo, space)
					fixes.append(gizmo)
			self.fixes = fixes
			self.old_blueprint = self.arch.my_blueprint
			if isinstance(self.arch, PlannedBase):
				self.arch._my_blueprint = self.spec


		def __exit__(self, exc_type, exc_val, exc_tb):
			for gizmo in self.fixes:
				self.arch.clear_space(gizmo)
			if isinstance(self.arch, PlannedBase):
				self.arch._my_blueprint = self.old_blueprint


	def build_with_spec(self, blueprint, *args, **kwargs):
		with self._Plan(self, blueprint):
			return self.build(*args, **kwargs)


	def validate_with_spec(self, blueprint, value):
		with self._Plan(self, blueprint):
			return self.validate(value)



#######################################################################################################################


# class AutoSpec(AbstractSpecced):
# 	_Spec = Spec
#
# 	def __init__(self, *args, blueprint=None, **kwargs):
# 		self._my_blueprint = blueprint
# 		super().__init__(*args, **kwargs)  # extracts hparams and processes crafts
# 		self.my_blueprint = self._update_spec(blueprint)
#
# 	@property
# 	def my_blueprint(self):
# 		return self._my_blueprint
#
# 	@my_blueprint.setter
# 	def my_blueprint(self, blueprint):
# 		self._my_blueprint = blueprint
#
# 	def _update_spec(self, spec=None):
# 		if spec is not None:
# 			return spec.update_with(self)
# # if spec is None:
# # 	spec = self._Spec()
# # if '_my_blueprint' not in self.__dict__:
# # 	return spec.update_with(self)
# # return spec
#
#
# class Specced(AutoSpec, ParameterizedBase, AbstractModular, SpaceKit):
# 	def __init__(self, *args, **kwargs):
# 		super().__init__(*args, **kwargs)  # extracts hparams
# 		self._fix_missing_spaces(self.my_blueprint)
# 		self._create_missing_submodules(self.my_blueprint)
#
# 	def _missing_spaces(self) -> Iterator[str]:
# 		for gizmo, skills in self._spaces.items():
# 			if len(skills) == 0:
# 				yield gizmo
# 			else:
# 				skill = skills[0]
# 				if skill.is_missing():
# 					yield gizmo
#
# 	def _fix_missing_spaces(self, spec):
# 		if spec is not None:
# 			for gizmo in self._missing_spaces():
# 				try:
# 					space = spec.space_of(gizmo)
# 				except ToolFailedError:
# 					continue
# 				else:
# 					self.change_space_of(gizmo, space)
#
# 	def _create_missing_submodules(self, spec):
# 		# if spec is not None:
# 		for name, param in self.named_submodules(hidden=True):
# 			try:
# 				val = getattr(self, name)
# 			except AttributeError:
# 				if spec is None:
# 					val = param.build_with(self)
# 				else:
# 					val = param.build_with_spec(self, spec.sub(name))
# 			else:
# 				val = param.validate(val, spec=spec)
# 			setattr(self, name, val)
#
# 		if spec is not None:
# 			spec.update_with(self)
#
#
# # def check_spec(self, spec):
# # 	# for name, param in self.named_hyperparameters(hidden=True):
# # 	# 	try:
# # 	# 		val = getattr(self, name)
# # 	# 	except AttributeError:
# # 	# 		if isinstance(param, AbstractSubmodule):
# # 	# 			val = param.build_with_spec(self, spec.sub(name))
# # 	# 			setattr(self, name, val)
# # 	# 	else:
# # 	# 		val = param.validate(val)
# # 	# 		setattr(self, name, val)
# # 	raise NotImplementedError
#
#
# class ArchitectBase(Specced, BuilderBase, AbstractArgumentBuilder):
# 	pass


########################################################################################################################


# class PlannedCrafty(ParamPlanned, IndividualCrafty): # comes before SpaceKit/Industrial
# 	# def _process_crafts(self):
# 	# 	for name, sub in self.named_submodules(hidden=True):
# 	# 		if sub.is_missing(self):
# 	# 			sub.setup_builder(self, spec=self.my_blueprint)
# 	# 	return super()._process_crafts()
#
# 	# def _process_skill(self, src: Type[AbstractCrafty], key: str, craft: AbstractCraft, skill: AbstractSkill):
# 	# 	super()._process_skill(src, key, craft, skill)
# 	pass
#
#
# # class PlannedBuilder(PlannedCrafty, BuilderBase):
# # 	def _integrate_spec(self, spec=None):
# # 		if spec is None:
# # 			spec = self.my_blueprint
# #
# # 		spec.node(self)  # add self to spec
# # 		for name, sub in self.named_submodules(hidden=True):
# # 			sub.setup_spec(self,
# # 			               spec=spec.sub(name))  # this should add builder/submodule to sub-spec (for space inference)
# #
# # 		for name in self._missing_spaces():  # explicitly requested spaces are inferred from spec (potentially using the subs)
# # 			self.change_space_of(name, spec.space_of(name))
# #
# # 		for name, space in self.named_spaces():
# # 			if space is None:
# # 				self.change_space_of(name, spec.space_of(name))
# #
# # 		for name, sub in self.named_submodules(hidden=True):
# # 			sub.validate(self)  # if sub is missing or a builder, this will instantiate it (using the sub-spec)
# #
# #
# # 	def space_of(self, gizmo: str) -> spaces.Dim:
# # 		try:
# # 			return super().space_of(gizmo)
# # 		except ToolFailedError:
# # 			return self.my_blueprint.space_of(gizmo)
# #
# # 	pass
#
#
#
# class PlannedBuilder(AbstractSpecced):
# 	def build(self, *args, blueprint=None, **kwargs):
# 		if blueprint is None:
# 			blueprint = self._my_blueprint
# 		return self._build(*args, blueprint=blueprint, **kwargs)
#
#
#
# class PlannedSpatial(PlannedCrafty, SpaceKit):
# 	def _coordinate_submodules(self, spec=None):
# 		if spec is None:
# 			spec = self.my_blueprint
#
# 		for name, sub in self.named_submodules(hidden=True):
# 			# this should add builder/submodule to sub-spec (for space inference)
# 			sub.setup_spec(self, spec=spec.sub(name))
#
# 		# explicitly requested spaces are inferred from spec and set to self (potentially using the subs)
# 		# add self to spec (implicitly adds all spaces defined in self to spec) (also means space wise self is ready)
# 		spec.certify(self)
#
# 		for name, sub in self.named_submodules(hidden=True):
# 			# if sub is missing or a builder, this will instantiate it (using the sub-spec)
# 			sub.validate(self)
#
#
# 	# def _fix_missing_spaces(self, spec):
# 	#
# 	# 	# find all missing spaces (in self + submodules)
# 	# 	todo = list(self._missing_spaces())
# 	#
# 	# 	# fill in missing spaces that are in blueprint
# 	#
# 	# 	# fill in remaining missing spaces using submodules
# 	#
# 	# 	# update blueprint with new spaces (and resolve conflicts)
# 	#
# 	# 	pass
#
#
#
# class CreativePlanned(PlannedSpatial):
# 	def _create_missing_submodules(self, spec):
# 		# convert submodule builders into submodules
#
# 		# send configs as needed
#
# 		# certify submodules with applications
#
# 		pass

















from .crafts import MachineCraft, ContextedCraft, SpacedCraft, OptionalCraft, DefaultCraft, LoggingCraft, \
	TensorCraft, SizeCraft, IndexCraft, SampleCraft, IndexSampleCraft, SpaceCraft, \
	TransformedSpaceCraft, SignatureCraft
from .kits import MaterialedCrafty, AssessibleCrafty, SignaturedCrafty, RelabeledCrafty, ElasticCrafty, SpaceKit
from .context import SizedContext, DynamicContext, ScopeBase, Cached, SeededContext, ScopedContext

from .assessments import SimpleSignature


class machine(ContextedCraft, MachineCraft, SpacedCraft):
	@classmethod
	def is_derivative(cls, obj):
		return isinstance(obj._base, machine)
	
	
	class Skill(MachineCraft.Skill, SpacedCraft.Skill):
		pass

	optional = OptionalCraft
	default = DefaultCraft



class indicator(machine, LoggingCraft):
	class Skill(machine.Skill, LoggingCraft.Skill):
		pass



class material(ContextedCraft, TensorCraft, SpacedCraft):
	@classmethod
	def is_derivative(cls, obj):
		return isinstance(obj._base, cls.from_size) or isinstance(obj._base, cls.from_indices) \
			or isinstance(obj._base, cls.next_sample) or isinstance(obj._base, cls.sample_from_index)
	
	
	class Skill(TensorCraft.Skill, SpacedCraft.Skill):
		pass

	from_size = SizeCraft
	from_indices = IndexCraft
	next_sample = SampleCraft
	sample_from_index = IndexSampleCraft



class space(SpaceCraft):
	pass



class Context(Cached, SizedContext, DynamicContext):
	def __init__(self, *args, tools=None, **kwargs):
		super().__init__(**kwargs)
		if tools is not None:
			self.include(*tools)



class Guru(Context):
	def __init__(self, *tools, **kwargs):
		super().__init__(tools=tools, **kwargs)



class Spatial(ElasticCrafty, RelabeledCrafty, SpaceKit):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._process_crafts()



class Industrial(Spatial, MaterialedCrafty, SignaturedCrafty, AssessibleCrafty):
	pass



class Signature(SimpleSignature):
	pass


# class Elastic(ElasticCrafty):









from typing import Callable, Union, Optional, Any, Iterator, Iterable, Hashable, Type, \
	Sequence, List, Tuple, Union, Mapping, Set, Dict, TypeVar

import inspect
import logging
from functools import lru_cache
from collections import OrderedDict, UserDict, namedtuple

from itertools import chain, product

from omnibelt import agnostic, unspecified_argument, filter_duplicates, \
	extract_function_signature, args2kwargs
from omnibelt.crafts import InheritableCrafty, NestableCraft, \
	AbstractCrafty, AbstractCraft, AbstractSkill



prt = logging.getLogger('omniply')






from omnibelt import agnosticmethod
# import torch

from omniplex.parameters import abstract
from omniplex.modules.models import Function



class Extractor(Function, abstract.Extractor):
	@agnosticmethod
	def extract(self, observation):
		return self(observation)



class Encoder(Extractor, abstract.Encoder):
	@agnosticmethod
	def encode(self, observation):
		return self(observation)



class Decoder(Function, abstract.Decoder):
	@agnosticmethod
	def decode(self, latent):
		return self(latent)



class Generator(Function, abstract.Generator): # TODO update
	@agnosticmethod
	def sample(self, *shape, gen=None):
		raise NotImplementedError



class Discriminator(Function, abstract.Discriminator):
	@agnosticmethod
	def judge(self, observation):
		return self(observation)



class Augmentation(Function, abstract.Augmentation):
	@agnosticmethod
	def augment(self, observation):
		return self(observation)



class Criterion(Function, abstract.Criterion):
	@agnosticmethod
	def compare(self, observation1, observation2):
		return self(observation1, observation2)



class Metric(Criterion, abstract.Metric): # obeys triangle inequality
	@agnosticmethod
	def distance(self, observation1, observation2):
		return self(observation1, observation2)



class PathCriterion(Criterion, abstract.PathCriterion):
	@agnosticmethod
	def compare_path(self, path1, path2):
		return self(path1, path2)



class Interpolator(Function, abstract.Interpolator):
	# returns N steps to get from start to finish ("evenly spaces", by default)
	@staticmethod
	def interpolate(start, end, N):
		start, end = start.unsqueeze(1), end.unsqueeze(1)
		progress = torch.linspace(0., 1., steps=N+2, device=start.device).view(1, N+2, *[1] * len(start.shape[2:]))
		return start + (end - start) * progress



class Estimator(Function, abstract.Estimator):
	def predict(self, observation):
		return self(observation)



class Invertible(Function, abstract.Invertible):
	def forward(self, observation):
		return self(observation)


	@agnosticmethod
	def inverse(self, observation):
		raise NotImplementedError



class Compressor(Function, abstract.Compressor):
	def compress(self, observation):
		return self(observation)


	@staticmethod
	def decompress(data):
		raise NotImplementedError



class Quantizer(Function, abstract.Quantizer):
	def quantize(self, observation): # generally "removes" noise
		return self(observation)


	@staticmethod
	def dequantize(observation): # generally adds noise
		raise NotImplementedError





from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable

import math
# import torch

from ..features import Prepared, ProgressBarred
from ..tools.abstract import AbstractScope, AbstractTool, AbstractResource
from ..tools.moguls import BatchMogul, IteratorMogul, SelectionMogul, LimitMogul, \
	EpochStatMogul, SimpleMogul, SeedingMogul

from .abstract import AbstractProgression, AbstractBatch
from .errors import BudgetExceeded, EpochEnd, UnknownSize
from .sources import Shufflable



class ProgressionBase(SimpleMogul, SeedingMogul, AbstractProgression):
	def __init__(self, source: Optional[AbstractTool] = None, **kwargs):
		super().__init__(**kwargs)
		self._source = None
		self._resource = None
		self._current_context = None
		self._sample_count = 0
		self._batch_count = 0
		if source is not None:
			self.set_source(source)


	def done(self) -> bool:
		return False


	def set_source(self, source: AbstractTool):
		self._source = source
		self._resource = self._as_resource(source)
		return self


	@property
	def source(self) -> AbstractTool:
		return self._source


	@property
	def current_batch(self) -> AbstractBatch:
		return self._current_context


	def current_context(self) -> AbstractBatch:
		if self._current_context is None:
			return self.create_batch()
		return self._current_context


	@property
	def sample_count(self) -> int:
		return self._sample_count
	@property
	def batch_count(self) -> int:
		return self._batch_count


	def _validate_context(self, ctx: AbstractBatch) -> AbstractBatch:
		ctx: AbstractBatch = super()._validate_context(ctx)
		self._sample_count += ctx.size
		self._batch_count += 1
		self._current_context = ctx
		ctx.set_progress(self)
		return ctx


	def create_batch(self, size=None) -> AbstractBatch:
		self.prepare()
		return self._create_context(size=size)



class BatchProgression(ProgressionBase):
	def __init__(self, batch_size: Optional[int] = None, **kwargs):
		super().__init__(**kwargs)
		self._batch_size = batch_size


	@property
	def batch_size(self):
		return self._batch_size


	def _create_context(self, *args, size=None, **kwargs):
		if size is None:
			size = self.batch_size
		return super()._create_context(*args, size=size, **kwargs)



class EpochProgression(ProgressionBase, EpochStatMogul):
	def __init__(self, epoch_size: Optional[int] = None, **kwargs):
		super().__init__(**kwargs)
		self._indices = None
		self._order_index = 0
		self._epoch_size = epoch_size

		self._batch_in_epoch = 0
		self._samples_in_epoch = 0


	def set_source(self, source: AbstractTool):
		self._indices = None
		return super().set_source(source)


	def _setup_epoch(self, remaining: Optional['torch.Tensor'] = None):
		self._indices = self._generate_sample_order()
		# TODO: add option to not overlap epochs (e.g. for contrastive learning, to make sure samples in batch are unique)
		if remaining is not None:
			self._indices = torch.cat([remaining, self._indices])
		self._order_index = 0
		self._batch_in_epoch = 0
		self._samples_in_epoch = 0


	_EpochEnd = EpochEnd
	def _reset_epoch(self):
		remaining = None
		if self._indices is not None:
			remaining = self.epoch_size - self._order_index
			remaining = self._indices[self._order_index:self._order_index + remaining]
		self._setup_epoch(remaining)


	def _validate_context(self, ctx: AbstractBatch) -> AbstractBatch:
		ctx = super()._validate_context(ctx)
		self._batch_in_epoch += 1
		self._samples_in_epoch += ctx.size
		return ctx


	def _next_batch_indices(self, size: int) -> 'torch.Tensor':
		remaining = self.epoch_size - self._order_index
		if remaining < size:
			self._reset_epoch()
			return self._next_batch_indices(size)
		indices = self._indices[self._order_index:self._order_index + size]
		self._order_index += size
		return indices


	def _create_context(self, *args, indices: Optional['torch.Tensor'] = None, size: Optional[int] = None, **kwargs):
		if self._indices is None:
			self._setup_epoch()

		if size is None:
			size = self.batch_size
		if indices is None:
			indices = self._next_batch_indices(size)

		return super()._create_context(*args, indices=indices, **kwargs)


	def _prepare(self, *args, **kwargs):
		super()._prepare(*args, **kwargs)
		assert self.batch_size <= self.epoch_size, f'Batch size ({self.batch_size}) must be ' \
		                                           f'less than or equal to epoch size ({self.epoch_size})'


	@property
	def epoch_size(self) -> int:
		if self._epoch_size is None:
			return self.source.size
		return self._epoch_size
	@epoch_size.setter
	def epoch_size(self, epoch_size: int):
		self._epoch_size = epoch_size


	@property
	def batch_in_epoch(self) -> int:
		return self._batch_in_epoch
	@property
	def samples_in_epoch(self) -> int:
		return self._samples_in_epoch


	def _generate_sample_order(self) -> 'torch.Tensor':
		return self.source._validate_selection(torch.arange(self.epoch_size))



class ShuffleProgression(EpochProgression, Shufflable):
	def __init__(self, *, shuffle: Optional[bool] = False, **kwargs):
		super().__init__(**kwargs)
		self._shuffle_batches = shuffle


	def _generate_sample_order(self) -> 'torch.Tensor':
		if self._shuffle_batches:
			return self._shuffle_indices(self.epoch_size)
		return super()._generate_sample_order()



#### top-level classes


class SetProgression(BatchProgression, EpochProgression):
	pass



class StreamProgression(BatchProgression):
	pass

















from typing import Dict, Tuple, Optional, Any, Callable, Type

import logging
from omnibelt import unspecified_argument, Class_Registry, agnostic, Modifiable, inject_modifiers, agnosticproperty
from omnibelt.tricks import auto_methods, dynamic_capture, extract_function_signature
import omnifig as fig

from ..tools.assessments import Signatured, AbstractSignature, SimpleSignature

from .abstract import AbstractBuilder, AbstractParameterized, AbstractArgumentBuilder, \
	AbstractMultiBuilder, AbstractRegistryBuilder
from .errors import NoProductFound, MissingBranchError
from .hyperparameters import HyperparameterBase
from .parameterized import ParameterizedBase
from ..structure import spaces

prt = logging.Logger('Building')
ch = logging.StreamHandler()
ch.setFormatter(logging.Formatter('%(levelname)s: %(msg)s'))
ch.setLevel(0)
prt.addHandler(ch)



class BuilderBase(ParameterizedBase, AbstractBuilder):
	pass




class AdaptableBuilderBase(ParameterizedBase, AbstractBuilder):
	def __call__(self, *args, **kwargs):
		return self.copy(*args, **kwargs)



class AnalysisBuilder(AbstractBuilder):
	_Signature = SimpleSignature
	def product_signatures(self, *args, **kwargs):
		product = self.product(*args, **kwargs)
		yield from product.signatures()



# class SelfAware(BuildableBase):
# 	pass
# # 	# TODO: maybe replace with SpecBuilder (or at least standardize behavior for stateful builders)
# #
# # 	def _populate_existing(self, target, existing=None):
# # 		if existing is None:
# # 			existing = {}
# # 		for name, _ in target.named_hyperparameters(hidden=True):
# # 			if name not in existing:
# # 				try:
# # 					value = getattr(self, name, unspecified_argument)
# # 				except self.Hyperparameter.MissingValueError:
# # 					continue
# # 				if value is not unspecified_argument:
# # 					if isinstance(value, SelfAware):
# # 						value = value.build_replica()
# # 					existing[name] = value
# # 		return existing
# #
# #
# # 	def build_replica(self, *args, **kwargs):
# # 		product = self.product(*args, **kwargs)
# # 		fixed_args, necessary_kwargs, missing = extract_function_signature(product, args=args, kwargs=kwargs,
# # 		                                                               allow_positional=False, include_missing=True,
# # 		                                                               default_fn=self._find_missing_hparam(self))
# # 		assert not len(missing), f'Could not find values for {missing} in {self}'
# # 		fixed_kwargs = self._populate_existing(product, necessary_kwargs)
# # 		return product(*fixed_args, **fixed_kwargs)



class ConfigBuilder(BuilderBase, AbstractArgumentBuilder, fig.Configurable):
	class _config_builder_type(fig.Configurable._config_builder_type):
		def __init__(self, product, config, *, target_name='__init__', silent=None):
			super().__init__(product, config, silent=silent)
			self.target_name = target_name


		def build(self, *args, **kwargs):
			init_capture = dynamic_capture(self.configurable_parents(self.product),
			                               self.fixer, self.target_name).activate()

			self.product._my_config = self.config

			target = self.product if self.target_name == '__init__' else getattr(self.product, self.target_name)
			obj = target(*args, **kwargs)

			del self.product._my_config
			obj._my_config = self.config

			init_capture.deactivate()
			return obj


	@classmethod
	def _config_builder(cls, config, silent=None, target_name='__init__'):
		return cls._config_builder_type(cls(), config, target_name=target_name, silent=silent)


	@classmethod
	def build_from_config(cls, config, args: Optional[Tuple] = None, kwargs: Optional[Dict[str, Any]] = None, *,
	                     silent: Optional[bool] = None) -> Any:
		if args is None:
			args = ()
		if kwargs is None:
			kwargs = {}
		return cls._config_builder(config, target_name='build', silent=silent).build(*args, **kwargs)


	def build(self, *args, **kwargs):
		product = self.product(*args, **kwargs)
		kwargs = self._build_kwargs(product, *args, **kwargs)
		config = self._my_config
		if config is not None and issubclass(product, fig.Configurable):
			return product.init_from_config(config, kwargs=kwargs)
		return product(**kwargs)



class ModifiableProduct(BuilderBase, AbstractArgumentBuilder):
	_product_mods = ()

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.__dict__['_product_mods'] = self._product_mods # save current product mods to instance.__dict__


	@staticmethod
	def _modify_product(product, *mods, name=None):
		if issubclass(product, Modifiable):
			return product.inject_mods(*mods, name=name)
		new = inject_modifiers(product, *mods, name=name)
		if new is not product and issubclass(new, ParameterizedBase):
			for src in reversed(new.__bases__):
				if issubclass(src, ParameterizedBase):
					new.inherit_hparams(*src.hyperparameter_names(hidden=True))
		return new


	def product_base(self, *args, **kwargs):
		raise NotImplementedError


	def modded(self, *mods):
		self._product_mods = [*self._product_mods, *mods]
		return self


	def vanilla(self):
		self._product_mods = ()
		return self


	def mods(self):
		yield from self._product_mods


	def product(self, *args, **kwargs) -> Type:
		return self._modify_product(self.product_base(*args, **kwargs), *self._product_mods)



# @fig.creator('build')
class BuildCreator(fig.Node.DefaultCreator): # creates using build() instead of __init__()
	@staticmethod
	def _modify_component(component, modifiers=()):
		component = component.cls
		if issubclass(component, ModifiableProduct):
			mods = [mod.cls for mod in modifiers]
			return component.modded(*mods)
		elif len(modifiers):
			raise NotImplementedError(f'Builders must subclass ModifiableProduct to use modifiers')
		return super()._modify_component(component, modifiers=modifiers)

	def _create_component(self, config, args: Tuple, kwargs: Dict[str, Any], silent: bool = None) -> Any:
		config.reporter.create_component(config, component_type=self.component_type, modifiers=self.modifiers,
		                                 creator_type=self._creator_name, silent=silent)

		cls = self._modify_component(self.component_entry,
		                             [self.project.find_artifact('modifier', mod) for mod in self.modifiers])

		if issubclass(cls, ConfigBuilder):
			obj = cls.build_from_config(config, args, kwargs, silent=silent)
		elif isinstance(cls, AbstractBuilder):
			settings = config.settings
			old_silent = settings.get('silent', None)
			settings['silent'] = silent
			obj = cls.build(*args, **kwargs)
			if old_silent is not None:
				settings['silent'] = old_silent
		else:
			raise NotImplementedError(f'Cannot create component of type {cls!r}')

		if isinstance(cls, ModifiableProduct):
			cls.vanilla()

		config._trace = None
		return obj



# class BuildableBase(ModifiableProduct, BuilderBase, AbstractArgumentBuilder):
# 	@classmethod
# 	def product_base(cls, *args, **kwargs) -> Type:
# 		return cls



# class AutoBuilder(BuilderBase, auto_methods, wrap_mro_until=True, # TODO: buildable should include __init__
#                   inheritable_auto_methods=['build', 'product', 'validate']):
#
# 	class MissingArgumentsError(TypeError):
# 		def __init__(self, src, method, missing, *, msg=None):
# 			if msg is None:
# 				msg = f'{src.__name__}.{method.__name__} missing {len(missing)} ' \
# 				      f'required arguments: {", ".join(map(str,missing))}'
# 			super().__init__(msg)
# 			self.missing = missing
# 			self.src = src
# 			self.method = method
#
# 	@classmethod
# 	def _auto_method_call(cls, self: Optional['AutoBuilder'], src: Type, method: Callable,
# 	                      args: Tuple, kwargs: Dict[str, Any]):
# 		# base = (cls if method.__name__ == '__init__' else src) if self is None else self
# 		base = cls if self is None else self
#
# 		fixed_args, fixed_kwargs, missing = extract_function_signature(method, args=args, kwargs=kwargs,
#                                                            allow_positional=False, include_missing=True,
#                                                            default_fn=base._find_missing_hparam(base))
# 		# fixed_args, fixed_kwargs, missing = extract_function_signature(method, args=args, kwargs=kwargs,
# 		#                                                                allow_positional=True, include_missing=True,
# 		#                                                                default_fn=base._find_missing_hparam(base))
# 		# # fixed_args, fixed_kwargs, missing = base.fill_hparams(method, args=args, kwargs=kwargs, include_missing=True)
# 		if len(missing):
# 			raise base.MissingArgumentsError(src, method, missing)
#
# 		return method(*fixed_args, **fixed_kwargs)



builder_registry = Class_Registry()
register_builder = builder_registry.get_decorator()
get_builder = builder_registry.get_class



class IdentBuilder(AbstractMultiBuilder, AbstractArgumentBuilder):
	_IdentParameter = HyperparameterBase

	def __init_subclass__(cls, _register_ident=True, default_ident=None, **kwargs):
		super().__init_subclass__(**kwargs)
		if _register_ident:
			cls.ident = cls._IdentParameter(required=True).setup(cls, 'ident')
		# cls._register_hparam('ident', cls._IdentParameter(name='ident', required=True), prepend=True)


	def __init__(self, ident=unspecified_argument, **kwargs):
		super().__init__(**kwargs)
		if ident is not unspecified_argument:
			self.ident = ident


	def build(self, ident: Optional[str] = unspecified_argument, **kwargs):
		if ident is unspecified_argument:
			ident = self.ident
		return super().build(ident=ident, **kwargs)


	def _build_kwargs(self, product, ident=unspecified_argument, **kwargs):
		if ident is unspecified_argument:
			ident = self.ident
		return super()._build_kwargs(product, ident=ident, **kwargs)



class MultiBuilderBase(IdentBuilder, ModifiableProduct):
	@classmethod
	def available_products(cls):
		return {}


	@classmethod
	def named_products(cls):
		yield from cls.available_products().items()


	def product_base(self, ident: Optional[str] = unspecified_argument, **kwargs):
		if ident is None:
			ident = self.ident
		product = self.available_products().get(ident, None)
		if product is None:
			raise self._NoProductFound(ident)
		return product



class RegistryBuilderBase(IdentBuilder, ModifiableProduct):
	class _IdentParameter(IdentBuilder._IdentParameter):
		_IdentSpace = spaces.Selection
		def __init__(self, *, space=None, **kwargs):
			if space is None:
				space = self._IdentSpace()
			super().__init__(space=space, **kwargs)


		def add_value(self, value, is_default=False):
			if value not in self.space:
				self.space.append(value)
			if is_default:
				self.default = value
			return self


	class _product_registry_type(Class_Registry, components=['owner']): pass
	_product_registry_owner = None
	_product_registry = None

	_default_products = None # these are inherited when set (unless overridden in __init_subclass__)
	_default_ident = None

	ident: _IdentParameter


	def __init_subclass__(cls, create_registry=None, default_ident=None, products=None, **kwargs):
		if create_registry is None:
			create_registry = cls._product_registry is None
		if default_ident is None:
			default_ident = cls._default_ident
		if products is None:
			products = {}
		if cls._default_products is not None:
			products.update(cls._default_products)
		super().__init_subclass__(_register_ident=create_registry, **kwargs)
		if create_registry:
			cls._product_registry = cls._product_registry_type()
			cls._product_registry_owner = cls
			for name, prod in products.items():
				cls.register_product(name, prod)

			if default_ident is not None:
				cls._register_ident(default_ident, is_default=True)


	@property
	def default_ident(self):
		return self.get_hparam('ident').default


	@classmethod
	def _register_ident(cls, ident, *, is_default=False):
		return cls.ident.add_value(ident, is_default=is_default)


	@classmethod
	def register_product(cls, name, product, *, is_default=False, **kwargs):
		registry = getattr(cls, '_product_registry', None)
		if registry is None:
			registry_name = cls.__name__ if isinstance(cls, type) else cls.__class__.__name__
			raise NotImplementedError(f'{registry_name} doesnt have a product registry')
		cls._register_ident(name, is_default=is_default)
		return registry.new(name, product, owner=cls, **kwargs)


	@classmethod
	def named_products(cls):
		for key, entry in cls._product_registry.items():
			yield key, entry.cls


	class _product_registration_decorator:
		def __init__(self, registry, name, **kwargs):
			self.registry = registry
			self.name = name
			self.kwargs = kwargs


		def __call__(self, product):
			self.registry.register_product(self.name, product, **self.kwargs)
			return product


	@classmethod
	def registration_decorator(cls, name, is_default=False, **kwargs):
		return cls._product_registration_decorator(cls, name, is_default=is_default, **kwargs)


	@classmethod
	def find_product_entry(cls, ident, default=unspecified_argument):
		entry = cls._product_registry.find(ident, None)
		if entry is None:
			if default is not unspecified_argument:
				return default
			raise cls._NoProductFound(ident)
		return entry


	def product_base(self, ident: Optional[str] = unspecified_argument, **kwargs):
		if ident is unspecified_argument:
			ident = self.ident
		entry = self.find_product_entry(ident)
		return entry.cls



class HierarchyBuilderBase(RegistryBuilderBase, create_registry=False):
	_branch_registry_type = Class_Registry
	_branch_registries = None
	_branch_name = None
	_branch_parent_registries = None
	_branch_address_delimiter = '/'
	_MissingBranchError = MissingBranchError


	def __init_subclass__(cls, branch=None, create_registry=None, parents=None, **kwargs):
		if branch is None and parents is not None:
			raise ValueError(f'{cls.__name__} is a branch but has no branch name')
		if branch is not None:
			if parents is None:
				parents = cls._product_registry_owner
			create_registry = create_registry is None or create_registry
		super().__init_subclass__(create_registry=create_registry, **kwargs)
		cls._branch_parent_registries = []
		cls._branch_registries = cls._branch_registry_type()

		if branch is not None:
			cls._branch_name = branch

		if branch is not None and parents is None:
			for base in cls.__bases__:
				if issubclass(base, HierarchyBuilderBase):
					base._branch_registries.new(branch, cls)
					# cls._branch_parent_registries.append(base)
					
		if parents is not None:
			assert branch is not None
			if isinstance(parents, (str, type)):
				parents = [parents]
			for owner in parents:
				if isinstance(owner, str):
					owner = get_builder(owner)
				if issubclass(owner, HierarchyBuilderBase):
					owner._branch_registries.new(branch, cls)
					# cls._branch_parent_registries.append(owner)


	@classmethod
	def hierarchy_registry_address(cls):
		if len(cls._branch_parent_registries) == 0:
			return cls._branch_name
		return f'{cls._branch_parent_registries[0].hierarchy_registry_address()}' \
		       f'{cls._branch_address_delimiter}{cls._branch_name}'


	@classmethod
	def branch_registries(cls):
		for branch, entry in cls._branch_registries.items():
			yield branch, entry.cls


	@classmethod
	def registry_hierarchy(cls):
		yield cls
		for entry in cls._branch_registries.values():
			yield from entry.cls.registry_hierarchy()


	@classmethod
	def product_hierarchy(cls):
		yield from cls.named_products()
		for entry in cls._branch_registries.values():
			for name, product in entry.cls.product_hierarchy():
				yield f'{entry.name}{cls._branch_address_delimiter}{name}', product


	@classmethod
	def find_product_entry(cls, ident, default=unspecified_argument):
		try:
			return super().find_product_entry(ident, default=default)
		except cls._NoProductFound:
			terms = ident.split(cls._branch_address_delimiter, 1)
			if len(terms) > 1:
				branch, ident = terms
				if branch in cls._branch_registries:
					return cls._branch_registries.get_class(branch).find_product_entry(ident, default=default)
				raise cls._MissingBranchError(branch, ident) from None
			raise



class RegisteredProductBase:
	ident = None
	_owning_registry = None

	def __init_subclass__(cls, ident=None, registry=None, is_default=False, **kwargs):
		super().__init_subclass__(**kwargs)
		if ident is None:
			ident = getattr(cls, 'ident', None)
		else:
			setattr(cls, 'ident', ident)
		if registry is None:
			registry = getattr(cls, '_owning_registry', None)
		else:
			if isinstance(registry, str):
				registry = get_builder(registry) # TODO: better error handling
			setattr(cls, '_owning_registry', registry)
		if ident is not None and registry is not None:
			registry.register_product(ident, cls, is_default=is_default)
			cls._owning_registry = registry



# class BoundRegistryBuilderBase(RegistryBuilderBase): # TODO: future
# 	_parent_registries = None
# 	_child_registries = None
#
# 	def __init_subclass__(cls, **kwargs):
# 		super().__init_subclass__(**kwargs)
# 		for base in cls.__bases__: # TODO: test extensively!
# 			if isinstance(base, RegistryBuilderBase) and base._product_registry is not None:
# 				cls._parent_registries.append(base._product_registry)
# 			if isinstance(base, BoundRegistryBuilderBase):
# 				base._child_registries.append(cls)


# class AutoRegistryBuilderBase(MultiBuilderBase):
# 	'''
# 	Issue differentiating a "builder" instance and a "product" instance. Specifically,
# 	which hparams are for the builder, and which ones are for the product.
# 	Conclusion: for now, separate the registration class from the registered product classes
#
# 	Example:
#
# 	```python
#
# 	class MyModels(ClassBuilder, nn.Module, default_ident='b'):
# 		p1 = hparam(required=True)
# 		p2 = hparam(10)
# 		p3 = hparam('hello', inherit=True)
# 		p4 = hparam((1,2,3), hidden=True)
#
#
# 	@inherit_hparams('p1', 'p2')
# 	class ModelB(MyModels, ident='b'):
# 		pass
#
# 	builder = MyModels(p2=50)
# 	assert MyModels.p2 == 10
# 	assert builder.p2 == 50
# 	b = builder.build()
# 	assert b.p2 == 50
# 	```
#
# 	'''
#
# 	Product_Registry = Class_Registry
# 	_product_registry: Product_Registry = None
# 	_registration_node = None
#
# 	def __init__(self, ident=None, **kwargs):
# 		if self.has_hparam('ident'): # self still behaves as a builder (bit of a hack)
# 			if ident is None:
# 				ident = self.ident
# 			super().__init__(ident=ident, **kwargs)
# 		else:
# 			super().__init__(**kwargs)
#
# 	class _IdentParameter(MultiBuilderBase.Hyperparameter):
# 		IdentSpace = spaces.Selection
#
# 		def __init__(self, *, space=None, **kwargs):
# 			if space is None:
# 				space = self.IdentSpace()
# 			super().__init__(space=space, **kwargs)
#
# 		def set_default_value(self, value):
# 			self.default = value
# 			if value not in self.space.values:
# 				self.space.values.append(value)
# 			return self
#
#
# 	def __init_subclass__(cls, create_registry=None, products=None, default_ident=unspecified_argument,
# 	                      _register_ident=None, **kwargs):
# 		if _register_ident is not None:
# 			prt.warning(f'`register_ident` should not be used with `RegistryBuilder`')
# 		prev_ident_hparam = None
# 		if create_registry is None and products is not None:
# 			create_registry = True
# 		if create_registry is None:
# 			create_registry = cls._registration_node is None
# 			prev_ident_hparam = cls.get_hparam('ident', None)
# 		super().__init_subclass__(_register_ident=create_registry, **kwargs)
# 		if create_registry:
# 			ident_hparam = cls.get_hparam('ident', None)
# 			if ident_hparam is not None and prev_ident_hparam is not None:
# 				ident_hparam.values.extend(prev_ident_hparam.values)
# 			cls._product_registry = cls.Product_Registry()
# 			cls._registration_node = cls
# 		if default_ident is not unspecified_argument:
# 			cls._set_default_product(default_ident)
# 		if products is not None:
# 			for name, product in products.items():
# 				cls.register_product(name, product)
#
#
# 	@classmethod
# 	def _set_default_product(cls, ident):
# 		return cls._registration_node.get_hparam('ident').set_default_value(ident)
#
# 	@classmethod
# 	def find_product_entry(cls, ident, default=unspecified_argument):
# 		entry = cls._registration_node._product_registry.find(ident, None)
# 		if entry is None:
# 			if default is not unspecified_argument:
# 				return default
# 			raise cls.NoProductFound(ident)
# 		return entry
#
# 	@classmethod
# 	def register_product(cls, name, product, is_default=False, **kwargs):
# 		cls._registration_node._product_registry.new(name, product, **kwargs)
# 		if is_default:
# 			cls._set_default_product(name)
#
# 	@classmethod
# 	def get_product_registration_decorator(cls):
# 		return cls._registration_node._product_registry.get_decorator()
#
#
# 	@agnostic
# 	def available_products(self):
# 		return {ident: entry.cls for ident, entry in self._registration_node._product_registry.items()}
#
# 	@agnostic
# 	def product(self, ident, **kwargs): # <------------------
# 		entry = self.find_product_entry(ident)
# 		return entry.cls
#
#
#
# class ClassBuilderBaseAuto(AutoRegistryBuilderBase, create_registry=False):
# 	'''Automatically register subclasses and add them to the product_registry.'''
#
# 	_class_builder_delimiter = '/'
# 	_class_builder_terms = ()
# 	def __init_subclass__(cls, ident=None, is_default=False, as_branch=False, **kwargs):
# 		super().__init_subclass__(**kwargs)
# 		if ident is not None:
# 			if as_branch:
# 				cls._class_builder_terms = (*cls._class_builder_terms, ident)
# 			else:
# 				addr = cls._class_builder_delimiter.join((*cls._class_builder_terms, ident))
# 				cls.ident = addr
# 				cls.register_product(addr, cls, is_default=is_default)










from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Generator, Type, Iterable, \
	Iterator, NamedTuple, ContextManager
import inspect
from omnibelt import split_dict, unspecified_argument, agnosticmethod, OrderedSet, \
	extract_function_signature, method_wrapper, agnostic, Modifiable

from ..persistent import AbstractFingerprinted, Fingerprinted
from ..tools.crafts import SpaceCraft

from .errors import InheritedHparamError
from .abstract import AbstractParameterized, AbstractHyperparameter
from .hyperparameters import InheritableHyperparameter



class ParameterizedBase(AbstractParameterized):
	def __init__(self, *args, **kwargs):
		params, remaining = self._extract_hparams(kwargs)
		super().__init__(*args, **remaining)
		for name, value in params.items():
			setattr(self, name, value)


	class _find_missing_hparam:
		def __init__(self, base, **kwargs):
			super().__init__(**kwargs)
			self.base = base


		def __call__(self, name, default=inspect.Parameter.empty):
			value = getattr(self.base, name, default)
			if value is inspect.Parameter.empty:
				raise KeyError(name)
			return value


	def fill_hparams(self, fn, args=None, kwargs=None, **finder_kwargs) -> Dict[str, Any]:
		params = extract_function_signature(fn, args=args, kwargs=kwargs, allow_positional=False,
		                                    default_fn=self._find_missing_hparam(self), **finder_kwargs)

		return params


	def _extract_hparams(self, kwargs):
		params = {}
		for name, _ in self.named_hyperparameters(hidden=True):
			if name in kwargs:
				params[name] = kwargs.pop(name)
				# setattr(self, name, kwargs.pop(name))
		return params, kwargs


	@classmethod
	def get_hparam(cls, key, default: Optional[Any] = unspecified_argument):
		# val = inspect.getattr_static(self, key, unspecified_argument)
		val = getattr(cls, key, unspecified_argument)
		if val is unspecified_argument:
			if default is unspecified_argument:
				raise AttributeError(f'{cls.__name__} has no hyperparameter {key}')
			return default
		return val


	def has_hparam(self, key):
		return isinstance(self.get_hparam(key, None), AbstractHyperparameter)


	@classmethod
	def named_hyperparameters(cls, *, hidden=False):
		items = list(cls.__dict__.items())
		for key, val in reversed(items):
			if isinstance(val, AbstractHyperparameter) and (hidden or not val.hidden):
				yield key, val



class InheritableParameterized(ParameterizedBase):
	_inherited_hyperparameters = None
	def __init_subclass__(cls, skip_inherit_inheritable_hparams=False, **kwargs):
		super().__init_subclass__(**kwargs)
		cls._inherited_hyperparameters = []
		if not skip_inherit_inheritable_hparams:
			existing = set(cls.hyperparameter_names(hidden=True))
			todo = []
			for base in cls.__bases__:
				if issubclass(base, ParameterizedBase):
					todo.extend(name for name, param in base.named_hyperparameters(hidden=True)
					            if name not in existing and isinstance(param, InheritableHyperparameter) and param.inherit)
			if len(todo):
				cls.inherit_hparams(*todo)


	@classmethod
	def named_hyperparameters(cls, *, hidden=False):
		yield from cls._inherited_named_hyperparameters(hidden=hidden)
		yield from super().named_hyperparameters(hidden=hidden)


	@classmethod
	def _inherited_named_hyperparameters(cls, *, hidden=False):
		if cls._inherited_hyperparameters is not None:
			for key in cls._inherited_hyperparameters:
				val = getattr(cls, key, None)
				if isinstance(val, AbstractHyperparameter) and (hidden or not val.hidden):
					yield key, val


	_InheritedHparamError = InheritedHparamError

	@classmethod
	def inherit_hparams(cls, *names):
		# for name in reversed(names):
		# 	val = getattr(cls, name, None)
		# 	if val is None:
		# 		raise cls._InheritedHparamError(f'{cls.__name__} cannot inherit the hparam: {name!r}')
		# 	setattr(cls, name, val)
		# return cls
		cls._inherited_hyperparameters.extend(name for name in names if name not in cls._inherited_hyperparameters)
		return cls



class ModifiableParameterized(ParameterizedBase, Modifiable):
	@classmethod
	def inject_mods(cls, *mods, name=None):
		product = super().inject_mods(*mods, name=name)
		product.inherit_hparams(*[key for src in [*reversed(mods), cls]
		                          for key, param in src.named_hyperparameters()])
		return product



class FingerprintedParameterized(ParameterizedBase, Fingerprinted):
	def _fingerprint_data(self):
		data = super()._fingerprint_data()
		hparams = {}
		for k, val in self.named_hyperparameters(hidden=True):
			try:
				hparams[k] = getattr(self, k)
			except AttributeError:
				pass
		data.update(hparams)
		return data



class SpatialParameterized(ParameterizedBase):
	def _extract_hparams(self, kwargs):
		params, remaining = super()._extract_hparams(kwargs)

		cls = type(self)
		for key in list(remaining.keys()):
			attr = getattr(cls, key, None) # TODO: maybe clean up a little somehow
			if isinstance(attr, SpaceCraft):
				params[key] = remaining[key]
				del remaining[key]

		return params, remaining



class InheritHparamsDecorator:
	def __init__(self, *names: Union[str, ParameterizedBase], **kwargs):
		self.names = names
		self.kwargs = kwargs


	_inherit_fn_name = 'inherit_hparams'

	def __call__(self, cls):
		try:
			inherit_fn = getattr(cls, self._inherit_fn_name)
		except AttributeError:
			raise TypeError(f'{cls} must be a subclass of {ParameterizedBase}')
		else:
			inherit_fn(*self.names, **self.kwargs)
		return cls



class HparamWrapper(method_wrapper):
	@staticmethod
	def process_args(args, kwargs, owner, instance, fn):
		base = owner if instance is None else instance
		return (), base.fill_hparams(fn, args, kwargs)
















from .kits import *



class Gig(AbstractGig):
	def _grab_from_fallback(self, error: GadgetFailedError, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		if ctx is None or ctx is self:
			raise AssemblyFailedError(gizmo, error)
			# raise error from error
		return ctx.grab(gizmo)


	def _grab(self, gizmo: str) -> Any:
		return super().grab_from(self, gizmo)


	def grab_from(self, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		try:
			out = self._grab(gizmo)
		except GadgetFailedError as error:
			out = self._grab_from_fallback(error, ctx, gizmo)
		return self.package(out, gizmo=gizmo)



class Cached(AbstractGig, UserDict):
	def __repr__(self):
		gizmos = [(gizmo if self.is_cached(gizmo) else '{' + gizmo + '}') for gizmo in self.gizmos()]
		return f'{self.__class__.__name__}({", ".join(gizmos)})'


	def gizmos(self) -> Iterator[str]:
		yield from filter_duplicates(self.cached(), super().gizmos())


	def is_cached(self, gizmo: str) -> bool:
		return gizmo in self.data


	def cached(self) -> Iterator[str]:
		yield from self.data.keys()


	def clear_cache(self) -> None:
		self.data.clear()


	# def _find_from(self, ctx: AbstractContext, gizmo: str) -> Any:
	# 	val = super().grab_from(ctx, gizmo)
	# 	self.data[gizmo] = val  # cache loaded val
	# 	return val


	def grab_from(self, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		if self.is_cached(gizmo):
			return self.data[gizmo]
		val = super().grab_from(ctx, gizmo)
		self.data[gizmo] = val  # cache loaded val
		return val



########################################################################################################################


class SimpleGroup(Gig, AbstractGroup):
	_current_context: Optional[AbstractGig]

	def __init__(self, *, apply: Optional[Dict[str, str]] = None, **kwargs):
		if apply is None:
			apply = {}
		super().__init__(**kwargs)
		self._raw_apply = apply
		self._raw_reverse_apply = None
		self._current_context = None


	@property
	def internal2external(self) -> Dict[str, str]:
		return self._raw_apply
	@internal2external.setter
	def internal2external(self, value: Dict[str, str]):
		self._raw_apply = value
		self._raw_reverse_apply = None


	@property
	def external2internal(self) -> Dict[str, str]:
		# return self._infer_external2internal(self._raw_apply, self.gizmoto())
		if self._raw_reverse_apply is None:
			self._raw_reverse_apply = self._infer_external2internal(self._raw_apply, self._gizmos())
		return self._raw_reverse_apply


	@staticmethod
	# @lru_cache(maxsize=None)
	def _infer_external2internal(raw: Dict[str, str], products: Iterator[str]) -> Dict[str, str]:
		reverse = {}

		for product in products:
			if product in raw:
				external = raw[product]
				if external in reverse:
					raise ApplicationAmbiguityError(product, [reverse[external], product])
				reverse[external] = product

		return reverse


	def gizmo_from(self, gizmo: str) -> str:
		return self.external2internal.get(gizmo, gizmo)


	def gizmo_to(self, gizmo: str) -> str:
		return self.internal2external.get(gizmo, gizmo)


	def _grab_from_fallback(self, error: GadgetFailedError, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		return super()._grab_from_fallback(error, self._current_context, self.gizmo_to(gizmo))


	def grab_from(self, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		if ctx is not None and ctx is not self:
			assert self._current_context is None, f'Context already set to {self._current_context}'
			self._current_context = ctx
			gizmo = self.gizmo_from(gizmo) # convert to internal gizmo
		return super().grab_from(self, gizmo)





from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable, Type, Set

import inspect
from functools import cached_property
from collections import OrderedDict

# import torch

from omnibelt import method_decorator, agnostic, unspecified_argument, filter_duplicates, get_printer
from omnibelt.crafts import AbstractCraft, AbstractCrafty, NestableCraft, SkilledCraft, IndividualCrafty


from .abstract import Loggable, AbstractAssessible, AbstractTool
from .errors import MachineError, ToolFailedError
from .assessments import Signatured

prt = get_printer(__name__)



class LabelCraft(SkilledCraft):
	class Skill(SkilledCraft.Skill):
		_base: 'LabelCraft'

		@property
		def label(self):
			return self.validate_label(self._base.label)


		def validate_label(self, label, *, owner = None):
			if owner is None:
				owner = self._instance
			return self._base._validate_label(owner, label)


	def __init__(self, label: str, **kwargs): # TODO: infer label from function name if not specified
		super().__init__(**kwargs)
		self._label = label


	def _validate_label(self, owner, label): # TODO: may be unnecessary (-> split into a subclass)
		return label


	@property
	def label(self):
		return self._label



class CopyCraft(LabelCraft):
	def copy(self, label: str = None, *args, **kwargs):
		# new = self.__class__(label, *args, **kwargs)
		# new.__dict__.update(self.__dict__)
		# return new
		if label is None:
			label = self.label
		return self.__class__(label, *args, **kwargs)


	def __copy__(self):
		return self.copy()



class ValidatedCraft(CopyCraft):
	def _validate_label(self, owner, label):  # TODO: may be unnecessary (-> split into a subclass)
		return owner.validate_label(label)



class ReplaceableCraft(ValidatedCraft):
	def __init__(self, label: str, *, replacements=None, **kwargs):
		if replacements is None:
			replacements = {}
		super().__init__(label, **kwargs)
		self.replacements = replacements

	@staticmethod
	def _update_application(old: Dict[str,str], new: Dict[str,str]): # TODO: test for infinite loops!
		existing = {}
		for key, value in old.items():
			existing.setdefault(value, []).append(key)

		composition = {}
		for k,v in new.items():
			if k in existing: # continue mapping
				for key in existing[k]:
					composition[key] = v
			else:
				composition[k] = v
		for k,v in old.items():
			if k not in composition:
				composition[k] = v
		return composition


	def replace(self, replacements: Dict[str, str], **kwargs):
		return self.copy(replacements=self._update_application(self.replacements, replacements), **kwargs)


	def _validate_label(self, owner, label):
		return super()._validate_label(owner, self.replacements.get(label, label))



class AnalysisCraft(ReplaceableCraft, Signatured, AbstractAssessible):
	class Skill(LabelCraft.Skill, Signatured, AbstractAssessible):
		_base: 'AnalysisCraft'


		def signatures(self, owner = None):
			if owner is None:
				owner = self._instance
			yield from self._base.signatures(owner)


	def signatures(self, owner: Type[AbstractCrafty] = None):
		yield self._Signature(self._validate_label(owner, self.label))



class ToolCraft(NestableCraft, AnalysisCraft):
	class Skill(AnalysisCraft.Skill, AbstractTool):
		_base: 'ToolCraft'

		def has_gizmo(self, gizmo: str) -> bool:
			return self._base._has_gizmo(self._instance, gizmo)


		def gizmos(self) -> Iterator[str]:
			yield from self._base._gizmos(self._instance)


		def get_from(self, ctx, gizmo: str):
			return self._base._get_from(self._instance, ctx, gizmo)


	def _has_gizmo(self, instance: AbstractCrafty, gizmo: str) -> bool:
		return gizmo == self._validate_label(instance, self.label)


	def _gizmos(self, instance: AbstractCrafty) -> Iterator[str]:
		yield self._validate_label(instance, self.label)


	def _get_from(self, instance: AbstractCrafty, ctx, gizmo: str):
		raise NotImplementedError



class LoggingCraft(ToolCraft):
	class Skill(ToolCraft.Skill, Loggable):
		_base: 'LoggingCraft'

		def log(self, ctx):
			return self._base._log(self._instance, ctx)


	def _log(self, instance: AbstractCrafty, ctx):
		raise NotImplementedError



class SimpleLoggingCraft(LoggingCraft):
	def log(self, instance: AbstractCrafty, ctx):
		return self._log_value(ctx[self.label])


	def _log_value(self, value):
		raise NotImplementedError



class FunctionCraft(method_decorator, NestableCraft, CopyCraft):
	def __init__(self, label: str, *, fn=None, **kwargs):
		super().__init__(label=label, fn=fn, **kwargs)


	def copy(self, label: str = None, *args, **kwargs):
		new = super().copy(label=label, *args, **kwargs)
		new._fn = self._fn
		new._name = self._name
		return new


	@property
	def wrapped(self):
		return self._fn


	_name = None
	def _setup_decorator(self, owner: Type, name: str) -> method_decorator:
		if self.label is not None and not self.label.isidentifier():
			prt.warning(f'Label {self.label} is not a valid identifier (for {owner.__name__}.{name})')
		self._name = name

		if self.wrapped is not None and isinstance(self.wrapped, method_decorator):
			self.wrapped._setup_decorator(owner, name)

		return super()._setup_decorator(owner, name)


	def setup(self, owner: Type, name: str):
		return self._setup_decorator(owner, name)


	def _get_instance_fn(self, instance: AbstractCrafty, name: Optional[str] = None):
		if name is None:
			name = self._name
		# if isinstance(instance, type):
		# 	content = self.content
		# 	return content
		fn = getattr(instance, name)
		if not hasattr(fn, '__name__'):
			if isinstance(fn, NestableCraft):
				return fn.content
			return self.content
		return fn



class FunctionToolCraft(FunctionCraft, ToolCraft):
	def _get_from(self, instance: AbstractCrafty, ctx, gizmo: str):
		return self._get_from_fn(instance, self._get_instance_fn(instance), ctx, gizmo)


	def _get_from_fn(self, instance, fn, ctx, gizmo):
		return fn(ctx)



class CustomArgumentCraft(FunctionToolCraft):
	def _get_fn_args(self, instance, fn, ctx, gizmo):
		return (), {}


	def _get_from_fn(self, instance, fn, ctx, gizmo):
		args, kwargs = self._get_fn_args(instance, fn, ctx, gizmo)
		return fn(*args, **kwargs)



class MetaArgCraft(CustomArgumentCraft):
	def _known_meta_args(self, owner) -> Iterator[str]:
		yield from ()


	def _known_input_args(self, owner) -> Iterator[str]:
		yield from ()


	def signatures(self, owner: Type[AbstractCrafty] = None):
		assert owner is not None
		label = self._validate_label(owner, self.label)
		yield self._Signature(label, inputs=tuple(self._known_input_args(owner)),
		                      meta=tuple(self._known_meta_args(owner)), fn=self._get_instance_fn(owner))



class SeededCraft(MetaArgCraft):
	_context_seed_key = 'seed'
	_context_rng_key = 'rng'

	def __init__(self, label: str = None, *, include_seed=None, include_rng=None, **kwargs):
		if include_seed and not isinstance(include_seed, str):
			include_seed = self._context_seed_key
		if include_rng and not isinstance(include_rng, str):
			include_rng = self._context_rng_key
		super().__init__(label=label, **kwargs)
		self._include_seed = include_seed
		self._include_rng = include_rng


	def _known_meta_args(self, instance):
		yield from super()._known_meta_args(instance)
		if self._include_seed:
			yield self._include_seed
		if self._include_rng:
			yield self._include_rng


	def _get_fn_args(self, instance, fn, ctx, gizmo):
		args, kwargs = super()._get_fn_args(instance, fn, ctx, gizmo)
		if self._include_seed:
			kwargs[self._include_seed] = getattr(ctx, self._context_seed_key, None)
		if self._include_rng:
			kwargs[self._include_rng] = getattr(ctx, self._context_rng_key, None)
		return args, kwargs



class AutoCraft(NestableCraft, LabelCraft):
	def __init__(self, label: Optional[str] = None, *args, **kwargs):
		super().__init__(label=label, *args, **kwargs)


	@property
	def label(self):
		if self._label is None:
			return self.content.__name__
		return self._label



class TransformableCraft(AbstractCraft):
	@staticmethod
	def _parse_fn_args(fn: Callable, *, raw: Optional[bool] = False,
	                   skip: Optional[Set[str]] = None) -> Iterator[Tuple[str, Any]]:
		params = inspect.signature(fn).parameters
		param_items = iter(params.items())
		if raw:
			next(param_items) # skip self/cls arg
		for name, param in param_items:
			if skip is None or name not in skip:
				yield name, param.default



class TransformCraft(TransformableCraft, MetaArgCraft):
	def _transform_inputs(self, owner, fn, *, raw: Optional[bool] = None):
		if raw is None:
			raw = isinstance(owner, type)
		skip = set(self._known_meta_args(owner))
		# for key, default in self._parse_fn_args(fn, raw=raw, skip=skip):
		# 	yield self._validate_label(owner, key), default
		yield from self._parse_fn_args(fn, raw=raw, skip=skip)


	_MachineError = MachineError
	def _fillin_fn_args(self, owner, fn: Callable, ctx, *, existing=None):
		# TODO: allow for arbitrary default values -> use omnibelt extract_signature

		if existing is None:
			existing = {}
		for key, default in self._transform_inputs(owner, fn, raw=False):
			name = self._validate_label(owner, key)
			try:
				existing[key] = ctx[name]
			except KeyError:
				if default is inspect.Parameter.empty:
					raise self._MachineError(key, self.label, owner)
				existing[key] = default

		return existing


	def _known_input_args(self, owner) -> Tuple:
		yield from filter_duplicates(super()._known_input_args(owner),
		        (self._validate_label(owner, key)
		         for key, default in self._transform_inputs(owner, self._get_instance_fn(owner))
		         if default is inspect.Parameter.empty))


	def _get_fn_args(self, instance, fn, ctx, gizmo):
		args, kwargs = super()._get_fn_args(instance, fn, ctx, gizmo)
		kwargs.update(self._fillin_fn_args(instance, fn, ctx))
		return args, kwargs



class SignatureCraft(TransformCraft):
	def __init__(self, signature, *, fn_name=None, label=None, **kwargs):
		if fn_name is None:
			assert signature.fn is not None, f'no function provided for signature {signature}'
			fn_name = signature.fn.__name__
		if label is None:
			label = signature.output
		else:
			signature.output = label
		super().__init__(label=label, **kwargs)
		self._fn_name = fn_name
		self._signature = signature


	_MachineError = MachineError
	def _fillin_fn_args(self, owner, fn: Callable, ctx, *, existing=None):
		# TODO: allow for arbitrary default values -> use omnibelt extract_signature
		if len(self._signature.meta):
			raise NotImplementedError # TODO

		inputs = [existing[key] if existing is not None and key in existing else ctx[key]
		          for key in self._known_input_args(owner)]
		return inputs


	def replace(self, replacements: Dict[str, str], **kwargs):
		new = super().replace(replacements, **kwargs)
		new._signature = self._signature.replace(replacements)
		return new


	def _known_meta_args(self, owner) -> Iterator[str]:
		yield from self._signature.meta


	def _known_input_args(self, owner) -> Iterator[str]:
		yield from self._signature.inputs


	def signatures(self, owner: Type[AbstractCrafty] = None):
		yield self._signature



class MachineCraft(AutoCraft, SeededCraft, TransformCraft):
	_MachineError = MachineError
	def _fillin_fn_args(self, owner, fn: Callable, ctx, *, existing=None):
		# TODO: allow for arbitrary default values -> use omnibelt extract_signature

		if existing is None:
			existing = {}
		for key, default in self._transform_inputs(owner, fn, raw=False):
			name = self._validate_label(owner, key)
			try:
				existing[key] = ctx[name]
			except KeyError:
				if default is inspect.Parameter.empty:
					raise self._MachineError(key, self.label, owner)
				existing[key] = default

		return existing


	def _get_fn_args(self, instance, fn, ctx, gizmo):
		args, kwargs = super()._get_fn_args(instance, fn, ctx, gizmo)
		kwargs.update(self._fillin_fn_args(instance, fn, ctx))
		return args, kwargs



class MaterialCraft(AutoCraft, SeededCraft):
	pass



class ContextToolCraft(MaterialCraft):
	def _get_fn_args(self, instance, fn, ctx, gizmo):
		args, kwargs = super()._get_fn_args(instance, fn, ctx, gizmo)
		return (ctx, *args), kwargs



class BatchCraft(MaterialCraft):
	_context_batch_key = 'batch'
	def _get_fn_args(self, instance, fn, ctx, gizmo):
		args, kwargs = super()._get_fn_args(instance, fn, ctx, gizmo)
		return (getattr(ctx, self._context_batch_key, ctx), *args), kwargs


	def _known_meta_args(self, instance):
		yield self._context_batch_key
		yield from super()._known_meta_args(instance)



class SizeCraft(MaterialCraft):
	_context_size_key = 'size'
	def _get_fn_args(self, instance, fn, ctx, gizmo):
		args, kwargs = super()._get_fn_args(instance, fn, ctx, gizmo)
		return (getattr(ctx, self._context_size_key, ctx), *args), kwargs


	def _known_meta_args(self, instance):
		yield self._context_size_key
		yield from super()._known_meta_args(instance)



class IndexCraft(MaterialCraft):
	_context_indices_key = 'indices'
	def _get_fn_args(self, instance, fn, ctx, gizmo):
		args, kwargs = super()._get_fn_args(instance, fn, ctx, gizmo)
		return (getattr(ctx, self._context_indices_key, ctx), *args), kwargs


	def _known_meta_args(self, instance):
		yield self._context_indices_key
		yield from super()._known_meta_args(instance)



class SampleCraft(MaterialCraft):
	_context_size_key = 'size'
	def _get_from_fn(self, instance, fn, ctx, gizmo):
		size = getattr(ctx, self._context_size_key, None)
		args, kwargs = self._get_fn_args(instance, fn, ctx, gizmo)
		if size is None:
			return fn(*args, **kwargs)
		return torch.stack([fn(*args, **kwargs) for _ in range(size)])



class IndexSampleCraft(MaterialCraft):
	_context_index_key = 'index'
	_context_indices_key = 'indices'


	def _get_from_fn(self, instance, fn, ctx, gizmo):
		indices = getattr(ctx, self._context_indices_key, None)
		args, kwargs = self._get_fn_args(instance, fn, ctx, gizmo)
		if indices is None:
			return fn(getattr(ctx, self._context_index_key, ctx), *args, **kwargs)
		return torch.stack([fn(i, *args, **kwargs) for i in indices])


	def _known_meta_args(self, instance):
		yield self._context_index_key
		yield from super()._known_meta_args(instance)



class CachedPropertyCraft(ReplaceableCraft, cached_property):
	def __init__(self, label: str, *, func=None, **kwargs):
		super().__init__(label=label, func=func, **kwargs)


	def copy(self, label: str = None, **kwargs):
		new = super().copy(label=label, **kwargs)
		new.func = self.func
		new.attrname = self.attrname
		return new


	def is_cached(self, instance: AbstractCrafty):
		return self.attrname in instance.__dict__


	def __call__(self, func: Callable):
		self.func = func
		return self


	def _get_instance_val(self, instance: AbstractCrafty, default: Optional[Any] = unspecified_argument):
		return getattr(instance, self.attrname) if default is unspecified_argument \
			else getattr(instance, self.attrname, default)


	def update_value(self, instance: AbstractCrafty, value):
		instance.__dict__[self.attrname] = value


	def __get__(self, instance, owner=None):
		if self.func is None and self.attrname not in instance.__dict__:
			raise ToolFailedError(f'No function or value for {self.label} in {instance}')
		return super().__get__(instance, owner)

	
	def __set__(self, instance, value):
		if instance is None:
			return
		self.update_value(instance, value)



class SpaceCraft(CachedPropertyCraft): # TransformCraft
	class Skill(CachedPropertyCraft.Skill):
		_base: 'SpaceCraft'


		def change_space_of(self, gizmo: str, space: str):
			return self._base._change_space_of(self._instance, gizmo, space)


		def space_of(self, gizmo: str):
			return self._base._space_of(self._instance, gizmo)


		def is_missing(self, gizmo: str = None):
			if gizmo is None:
				gizmo = self.label
			return self._base._is_missing(self._instance, gizmo)


		def clear_space(self, gizmo: str) -> bool:
			return self._base._clear_space(self._instance, gizmo)

	
	def _change_space_of(self, instance: AbstractCrafty, gizmo, space):
		self.update_value(instance, space)


	def _space_of(self, instance: AbstractCrafty, gizmo: str = None):
		return self._get_instance_val(instance)


	def _is_missing(self, instance: AbstractCrafty, gizmo: str = None):
		return self.func is None and self.attrname not in instance.__dict__


	def _clear_space(self, instance: AbstractCrafty, gizmo: str = None):
		if self.attrname in instance.__dict__:
			del instance.__dict__[self.attrname]
			return True
		return False



class TransformedSpaceCraft(TransformableCraft, SpaceCraft): # TODO: future
	def _get_instance_val(self, instance: AbstractCrafty, default: Optional[Any] = unspecified_argument):
		fn = self.func
		if self.is_cached(instance) or fn is None \
				or len(inspect.signature(fn.__get__(instance, type(instance))).parameters) == 0:
			return super()._get_instance_val(instance, default)

		op = fn.__get__(instance, type(instance))
		bases = {}
		for name, val in self._parse_fn_args(op):
			bases[name] = instance.space_of(val) # TODO: deal with cycles

		return op(**bases)



# class SpecSpaceCraft(SpaceCraft):
# 		_missing_space_flag = object()
# 		def _space_of(self, instance: AbstractCrafty, gizmo: str = None):
# 			space = self._get_instance_val(instance, self._missing_space_flag)
# 			if space is self._missing_space_flag:
# 				return instance._default_space_of(gizmo)
# 			return space



# class SchemaSpaceCraft(SpaceCraft):
# 	_missing_space_flag = object()
# 	def _space_of(self, instance: AbstractCrafty, gizmo: str = None):
# 		space = self._get_instance_val(instance, self._missing_space_flag)
# 		if space is self._missing_space_flag:
# 			return instance._default_space_of(gizmo)
# 		return space



class ContextualSpaceCraft(SpaceCraft):
	pass



class OptionalCraft(MachineCraft):
	class Skill(MachineCraft.Skill):
		pass



class DefaultCraft(MachineCraft):
	class Skill(MachineCraft.Skill):
		pass



class InitCraft(CachedPropertyCraft):
	class Skill(CachedPropertyCraft.Skill):
		_base: 'InitCraft'


		def init(self, instance: Optional[AbstractCrafty] = None):
			if instance is None:
				instance = self._instance
			return self._base.init(instance)


	def init(self, instance: AbstractCrafty):
		return self._process_init_val(instance, self._get_instance_val(instance))


	def _process_init_val(self, instance, val):
		return val



class TensorCraft(InitCraft):
	def _process_init_val(self, instance, val):
		if isinstance(val, torch.Tensor):
			buffer = getattr(instance, 'Buffer', None)
			return buffer(val)
		return val



########################################################################################################################



class MethodCraft(AbstractCraft):
	'''For crafts which contain methods to create crafts, but can't be used directly themselves'''
	def __init__(self, *args, **kwargs):
		raise TypeError('This craft cannot be directly instantiated')



class SpacedCraft(LabelCraft):
	_space_craft_type = SpaceCraft

	@agnostic
	def space(self, *args, **kwargs):
		if isinstance(self, type):
			self: Type['SpacedCraft']
			return self._space_craft_type(*args, **kwargs)
		return self._space_craft_type(self.label)(*args, **kwargs)



class ContextedCraft(LabelCraft):
	from_context = ContextToolCraft
	from_batch = BatchCraft









from .imports import *
from .abstract import AbstractQuirk, T



class MissingValueError(KeyError):
	pass


class DescriptorQuirk(AbstractQuirk):
	def __init__(self, *, attrname: Optional[str] = None, owner: Optional[Type[T]] = None, **kwargs):
		super().__init__(**kwargs)
		self._attrname = attrname
		self._owner = owner


	def _replicator_kwargs(self, kwargs: dict[str, Any]) -> dict[str, Any]:
		if 'attrname' not in kwargs:
			kwargs['attrname'] = self._attrname
		return kwargs


	def __str__(self):
		return f'<quirk:{self._attrname}>'


	def __repr__(self):
		return f'{self.__class__.__qualname__}({self._attrname})'


	def __set_name__(self, owner: Type[T], name: str):
		if self._attrname is None:
			self._attrname = name
		elif name != self._attrname:
			raise TypeError(
				f"Cannot assign the same {self.__class__.__qualname__} to two different names "
				f"({self._attrname!r} and {name!r})."
			)
		self._owner = owner
		return self


	def __get__(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		if self._getter is None:
			return self.realize(instance, owner=owner)
		if instance is None:
			raise self._MissingOwnerError(self)
		return self._getter.__get__(instance, owner=owner)()


	def __set__(self, instance: T, value: Any):
		self.reset(instance, value)


	def __delete__(self, instance: T):
		self.rip(instance)



class DefaultQuirk(DescriptorQuirk):
	_missing_value = object()

	def __init__(self, default: Optional[Any] = unspecified_argument, *,
	             getter: Optional[Callable[[T], Any]] = None, **kwargs):
		if default is unspecified_argument:
			default = self._missing_value
		getter, default = self._check_fget(getter, default)
		super().__init__(**kwargs)
		self._default = default
		self._getter = getter


	_MissingValueError = MissingValueError
	def realize(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		if self._default is self._missing_value:
			raise self._MissingValueError(self)
		return self._default


	def __get__(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		if self._getter is None:
			return self.realize(instance, owner=owner)
		if instance is None:
			raise self._MissingOwnerError(self)
		return self._getter.__get__(instance, owner=owner)()

	pass














from tqdm import tqdm

from omnibelt import unspecified_argument, agnosticproperty

from .errors import NotReady


class Named:
	_name = None


	class NoNameError(ValueError):
		pass


	def __str__(self):
		if self.name is None:
			raise self.NoNameError('No name specified')
		return self.name


	@agnosticproperty
	def name(self):
		return getattr(self, '_name', None)
	@name.setter
	def name(self, name):
		self._name = name



class InitWall:
	def __init__(self, *args, **kwargs):
		if len(args) > 0:
			raise TypeError(f'Unexpected args: {args}')
		if len(kwargs) > 0:
			raise TypeError(f'Unexpected kwargs: {", ".join(kwargs.keys())}')
		super().__init__()



class Prepared: # TODO: add autoprepare using __certify__
	_NotReady = NotReady


	@property
	def is_ready(self):
		return getattr(self, '_is_ready', False)


	def prepare(self, *args, **kwargs):
		if not self.is_ready:
			self._is_ready = True
			self._prepare(*args, **kwargs)
		return self


	def _prepare(self, *args, **kwargs):
		pass



class ProgressBarred:
	_pbar_owner = None
	_pbar_cls = tqdm
	_pbar_instance = None


	@classmethod
	def replace_pbar_cls(cls, pbar):
		cls._pbar_owner._pbar_cls = pbar


	@classmethod
	def _create_pbar(cls, *args, **kwargs):
		if cls._pbar is not None:
			cls._close_pbar()
		cls._pbar_owner._pbar = cls._pbar_cls(*args, **kwargs)
		return cls._pbar


	@agnosticproperty
	def _pbar(self):
		return self._pbar_owner._pbar_instance


	@classmethod
	def _close_pbar(cls):
		if cls._pbar is not None:
			cls._pbar.close()
			cls._pbar = None


ProgressBarred._pbar_owner = ProgressBarred









from .quirks import *
from .kits import *
from .contexts import *


class AbstractQuality(AbstractQuirk):
	# def realize(self, instance: T, owner: Type[T]) -> Any:
	# 	'''creates the quirk ab initio for the instance (usually when no default value is provided)'''
	# 	raise NotImplementedError

	_ReachFailedFlag = ReachFailedFlag
	def reach(self, owner: Optional[Type[T]] = None) -> Any:
		'''tries accessing the quirk without creating it (eg. for static analysis)'''
		raise self._ReachFailedFlag



class AbstractTrait(AbstractQuality, AbstractCraft):
	pass



class TemplateQuality(DefaultQuirk, AbstractQuality):
	def __init__(self, default: Optional[Any] = unspecified_argument, *,
	             tmpl: Optional[Type] = None, **kwargs):
		super().__init__(default, **kwargs)
		self._template = tmpl


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'tmpl' not in kwargs:
			kwargs['tmpl'] = self._template
		return kwargs


	def reach(self, owner: Optional[Union[T, Type[T]]] = None) -> Any:
		return self._template


	def realize(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		if self._default is self._missing_value:
			return self.reach(instance)()
		return self._default



class ArtQuality(DefaultQuirk, AbstractQuality):
	'''caches the artist in the instance's __dict__ using _builder_key() as key'''
	def __init__(self, default: Optional[Any] = unspecified_argument, *,
	             artist: Optional['AbstractArtist'] = None, **kwargs):
		super().__init__(default, **kwargs)
		self._artist = artist


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'artist' not in kwargs:
			kwargs['artist'] = self._artist
		return kwargs


	def reach(self, owner: Optional[Union[T, Type[T]]] = None) -> Any:
		'''creates a new artist'''
		return self._create_builder(owner) # TODO: maybe consolidate


	def realize(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		if self._default is self._missing_value:
			return self.reach(instance).design_for(instance)
		return self._default


	def _create_builder(self, owner: Optional[Union[T, Type[T]]] = None) -> 'AbstractArtist':
		if self._artist is None:
			raise self._MissingValueError(self)

		raise NotImplementedError  # TODO: implement this



class CachedArtQuality(ArtQuality):
	def _builder_key(self, owner: Optional[Union[T, Type[T]]] = None):
		return f'__builder_for_{self._attrname}'


	def realize(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		product = super().realize(instance, owner=owner)
		# remove cached artist after realizing the trait
		key = self._builder_key(owner)
		if hasattr(owner, key):
			delattr(owner, key)
		return product


	def reach(self, owner: Optional[Union[T, Type[T]]] = None) -> Any:
		if owner is None: # caching not possible (eg. static analysis)
			return super().reach(owner)

		key = self._builder_key(owner)
		artist = getattr(owner, key, None) # TODO: consider accessing __dict__ directly
		if artist is not None:
			return artist

		artist = self._create_builder(owner)
		setattr(owner, key, artist)
		return artist



class TraitTool(AbstractGadget): # TODO: should this be an AbstractMultiTool?
	'''
	delegates work to submodule by accessing it with `getattr(self.owner, self.attrname)`
	(thereby building lazily)
	'''
	def __init__(self, owner: AbstractCrafty, attrname: str, **kwargs):
		super().__init__(**kwargs)
		self._owner = owner
		self._attrname = attrname


	def _resolve_tool(self):
		return getattr(self._owner, self._attrname)


	def gizmos(self):
		return self._resolve_tool().gizmos()


	def grab_from(self, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		tool = self._resolve_tool()
		return tool.grab_from(ctx, gizmo)


	def __repr__(self):
		return f'[{self._attrname}]({", ".join(self.gizmos())})'



class ScopeTraitKit(TraitTool):
	def grab_from(self, ctx: 'AbstractScopedContext', gizmo: str) -> Any:
		return self._resolve_tool().grab_from(ctx.scope_for(self), gizmo)



class SimpleTrait(DescriptorQuirk, AbstractTrait):
	_Skill = TraitTool
	def as_skill(self, owner: AbstractCrafty, **kwargs):
		return self._Skill(owner=owner, attrname=self._attrname, **kwargs)



class AppliedTraitTool(SimpleGroup, TraitTool, AbstractMultiTool): # TODO: should this be an AbstractMultiTool?
	'''
	delegates work to submodule by accessing it with `getattr(self.owner, self.attrname)`
	(thereby building lazily)
	'''



class AppliedTrait(SimpleTrait):
	# submachine -> appliedtrait
	def __init__(self, *, apply: Optional[Dict[str, str]] = None, **kwargs):
		# apply is the mapping the dev specifies to map the trait's gizmos to the owner's gizmos
		super().__init__(**kwargs)
		self._apply = apply


	_Skill = AppliedTraitTool
	def as_skill(self, owner: AbstractCrafty, apply=unspecified_argument, **kwargs):
		if apply is unspecified_argument:
			apply = self._apply
		return super().as_skill(owner=owner, apply=apply, **kwargs)













from .imports import *



class Replicator:
	def _replicator_kwargs(self, kwargs: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
		if kwargs is None:
			kwargs = {}
		return kwargs


	def replicate(self, *args, **kwargs) -> Any:
		kwargs = args2kwargs(self.__class__, args, kwargs)
		return self.__class__(**self._replicator_kwargs(kwargs))










from .abstract import AbstractGizmo


class DashGizmo(AbstractGizmo):
	"""
	The DashGizmo class is a subclass of AbstractGizmo. It provides methods to handle gizmos with dashes.

	Attributes:
		_native (str): The original label of the gizmo.
		_fixed (str): The label of the gizmo with dashes replaced by underscores.
	"""

	__slots__ = ('_native', '_fixed')

	def __init__(self, label: str):
		"""
		Initializes a new instance of the DashGizmo class.

		Args:
			label (str): The label of the gizmo.
		"""
		self._native = label
		self._fixed = label.replace('-', '_')

	def __eq__(self, other):
		"""
		Checks if the DashGizmo instance is equal to another object.

		Args:
			other: The object to compare with.

		Returns:
			bool: True if the DashGizmo instance is equal to the other object, False otherwise.
		"""
		if isinstance(other, str):
			return str(self) == other.replace('-', '_')
		return str(self) == str(other)

	def __hash__(self):
		"""
		Returns the hash of the DashGizmo instance.

		Returns:
			int: The hash of the DashGizmo instance.
		"""
		return hash(str(self))

	def __str__(self):
		"""
		Returns a string representation of the DashGizmo instance.

		Returns:
			str: A string representation of the DashGizmo instance.
		"""
		return self._fixed















from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Hashable, Iterator, Iterable, Type, Set
from collections import UserDict
from omnibelt import unspecified_argument, filter_duplicates

from ..structure import spaces
from ..features import Seeded

from .abstract import AbstractContext, AbstractSpaced, AbstractKit, AbstractTool, AbstractAssessible, AbstractScope, \
	AbstractAssessment, AbstractSourcedKit, AbstractMogul, AbstractSchema, AbstractResource, AbstractDynamicContext
from .errors import MissingGizmoError, ToolFailedError
from .assessments import Signatured



class ContextBase(AbstractContext, AbstractSourcedKit, AbstractSpaced, AbstractAssessible, Signatured):
	def context_id(self) -> Hashable:
		return id(self)


	def __str__(self):
		return f'{self.__class__.__name__}({", ".join(self.gizmos())})'


	def vendors(self, gizmo: str) -> Iterator['AbstractTool']:
		for source in self.sources():
			if source.has_gizmo(gizmo):
				yield source


	def _get_from(self, ctx: AbstractContext, gizmo: str):
		for tool in self.vendors(gizmo):
			try:
				return tool.get_from(self, gizmo)
			except ToolFailedError:
				pass
		raise MissingGizmoError(gizmo)


	def space_of(self, gizmo: str) -> spaces.Dim:
		for tool in self.vendors(gizmo):
			try:
				return tool.space_of(gizmo)
			except (AttributeError, ToolFailedError):
				pass


	def get_from(self, ctx: Optional['AbstractContext'], gizmo: str):
		return self._get_from(ctx, gizmo)


	def assess(self, assessment: AbstractAssessment):
		super().assess(assessment)
		for source in self.sources():
			if isinstance(source, AbstractAssessible):
				assessment.add_edge(self, source)
				assessment.expand(source)


	def signatures(self, owner = None) -> Iterator['AbstractSignature']:
		for source in self.sources():
			if isinstance(source, Signatured):
				yield from source.signatures(self)



class ScopedContext(ContextBase):
	def __init__(self, *, scope_table=None, **kwargs):
		super().__init__(**kwargs)
		self._scope_table = scope_table or {}


	def clear_scopes(self):
		self._scope_table.clear()


	def scope_for(self, code, default=unspecified_argument):
		try:
			return self._scope_table[code]
		except KeyError:
			if default is unspecified_argument:
				raise
			return default


	def register_scope(self, code, scope):
		self._scope_table[code] = scope



class NestedContext(ContextBase):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._trace = []


	def _add_trace(self, ctx: AbstractContext):
		self._trace.append(ctx)
	def _pop_trace(self, ctx: AbstractContext):
		self._trace.pop()


	def _fallback_get_from(self, gizmo: str):
		for src in reversed(self._trace):
			try:
				return src.get_from(self, gizmo)
			except MissingGizmoError:
				continue
		raise MissingGizmoError(gizmo)


	def get_from(self, ctx: Optional['AbstractContext'], gizmo: str):
		if ctx is not self:
			self._add_trace(ctx)

		try:
			val = self._get_from(ctx, gizmo)
		except MissingGizmoError:
			val = self._fallback_get_from(gizmo)

		if ctx is not self:
			self._pop_trace(ctx)

		return val



class ScopeBase(NestedContext, AbstractScope):
	'''
	interface between the internal labels (defined by dev) for a single module,
	and the external labels (defined by the user) for the entire system
	'''
	# def gizmoto(self) -> Iterator[str]: # no mapping
	# 	for gizmo in self._base.gizmos():
	# 		yield self.gizmo_to()
	# 	yield from self.gizmos()


	def _fallback_get_from(self, gizmo: str):
		return super()._fallback_get_from(self.gizmo_from(gizmo))


	def get_from(self, ctx: Optional['AbstractContext'], gizmo: str):
		if ctx is not self:
			gizmo = self.gizmo_to(gizmo)
		return super().get_from(ctx, gizmo)



class ApplicationScope(AbstractScope):
	def __init__(self, application: Dict[str,str] = None,  **kwargs):
		if application is None:
			application = {}
		super().__init__(**kwargs)
		self._application = application
		self._reverse_application = {v:k for k,v in application.items()}


	def gizmo_from(self, gizmo: str) -> str:
		return self._reverse_application.get(gizmo, gizmo)


	def gizmo_to(self, external: str) -> str:
		return self._application.get(external, external)



class DynamicContext(ContextBase, AbstractDynamicContext):
	def __init__(self, *, sources=None, **kwargs):
		if sources is None:
			sources = []
		super().__init__(**kwargs)
		prev = []
		if len(sources):
			prev = list(sources)
			sources.clear()
		self._sources = sources
		self.include(*prev)


	def include(self, *sources: AbstractTool): # a source can be a scope
		self._sources.extend(reversed(sources))
		return self


	def sources(self): # N-O
		yield from reversed(self._sources)



class SizedContext(AbstractContext):
	def __init__(self, *args, size=None, **kwargs):
		super().__init__(**kwargs)
		self._size = size


	@property
	def size(self):
		return self._size



class SeededContext(ContextBase, Seeded):
	def _get_from(self, ctx, gizmo):
		with self.default_rng():
			return super()._get_from(ctx, gizmo)



class Cached(SeededContext, UserDict):
	def gizmos(self) -> Iterator[str]:
		yield from filter_duplicates(super().gizmos(), self.cached())


	def is_cached(self, gizmo: str):
		return gizmo in self.data


	def cached(self):
		yield from self.data.keys()


	def clear_cache(self):
		self.data.clear()


	def _get_from(self, ctx, gizmo):
		if self.is_cached(gizmo):
			val = self.data[gizmo]
		else:
			val = super()._get_from(self, gizmo)
			self.data[gizmo] = val # cache loaded val
		if ctx is self:
			return val
		return val[ctx.indices]


	def __str__(self):
		gizmos = [(gizmo if self.is_cached(gizmo) else '{' + gizmo + '}') for gizmo in self.gizmos()]
		return f'{self.__class__.__name__}({", ".join(gizmos)})'



class SimpleScope(ApplicationScope, Cached, DynamicContext):
	def __init__(self, source, **kwargs):
		super().__init__(**kwargs)
		self.include(source)





##########################################################################################

# class Decoder(Function):
# 	@machine('out')
# 	def forward(self, inp):
# 		# do something
# 		return out
#
#
#
# class Autoencoder2:
# 	encoder = submachine(builder='encoder', input='observation', output='latent')
# 	decoder = submachine(builder='decoder', input='latent', output='reconstruction')
#
# 	@machine('loss')
# 	def compute_loss(self, observation, reconstruction):
# 		return self.criterion(reconstruction, observation)





















from typing import Type, Union, Any, Optional, Callable, Sequence, Iterable, Iterator, Tuple, List, Dict, NamedTuple
from collections import OrderedDict
from omnibelt import smartproperty, unspecified_argument
# from omnibelt.tricks import nested_method_decorator
from omnibelt.collectors import method_propagator, universal_propagator, AbstractCollector, AbstractCollectorTrigger




class Container(OrderedDict): # contains gizmos
	class MissingKey(KeyError):
		pass
	
	
	def _find_missing(self, key):
		raise self.MissingKey(key)


	def __getitem__(self, item):
		try:
			return super().__getitem__(item)
		except KeyError:
			return self._find_missing(item)


	def __str__(self):
		entries = ', '.join(self.keys())
		return f'{self.__class__.__name__}({entries})'


	def __repr__(self):
		return str(self)


class AbstractDepot:
	pass


class Depot(Container): # contains machines
	def __init__(self, *sources, **kwargs):
		super().__init__(**kwargs)
		self._sources = sources


	def _load_missing(self, key):
		if key in self._gizmos:
			return self._gizmos[key].get_from(self, key)
		return self.source[key]
	
	
	def _find_missing(self, key):
		if key in self._gizmos:
			self[key] = self._gizmos[key].get_from(self, key)
			return self[key]
		if self._gizmos is not None:
			self[key] = self._load_missing(key) # load and cache
			return self[key]
		return super()._find_missing(key)



class machine_base(method_propagator):
	def gizmos(self):
		yield from self._keys


class AbstractMachine:
	def gizmos(self):
		raise NotImplementedError



class AbstractMachineTrigger:
	pass



class _Machine_Trigger(AbstractMachineTrigger):
	def __init__(self, owner, key, base):
		self.owner = owner
		self.key = key
		self.base = base



class MachineBase(AbstractMachine):
	_known_machines_type = list

	@classmethod
	def process_machine_triggers(cls, owner: Type['MachinedBase']):
		machines = []
		for base in owner.__bases__:
			if issubclass(base, MachinedBase):
				past = getattr(base, '_known_machines', None)
				if past is not None:
					machines.extend(past)
		for key, val in owner.__dict__.items():
			if isinstance(val, machine_base):
				machines.append(owner._Machine_Trigger(owner, key, val))

		known = set()
		used = cls._known_machines_type()
		for trigger in reversed(machines):
			gizmos = list(cls.extract_gizmos_from_trigger(trigger))
			if all(g in known for g in gizmos):
				continue
			known.update(gizmos)
			used.append(trigger)
		return used


	@staticmethod
	def extract_gizmos_from_trigger(trigger: 'AbstractMachineTrigger'):
		yield from trigger.val.gizmos()
		
		
	def gizmos(self):
		yield from self.extract_gizmos_from_trigger(self.trigger)


	_known_machines_gizmos_type = OrderedDict

	@classmethod
	def process_gizmos(cls, source, triggers):
		table = cls._known_machines_gizmos_type()
		for trigger in triggers:
			machine = cls(source, trigger)
			for gizmo in machine.gizmos():
				table[gizmo] = machine
		return table


	def __init__(self, source, trigger: 'AbstractMachineTrigger', **kwargs):
		super().__init__(**kwargs)
		self.source = source
		self.trigger = trigger



class MachinedBase:
	Machine = MachineBase
	_Machine_Trigger = _Machine_Trigger
	Depot = Depot

	_known_gizmos = None
	_known_machine_triggers = None
	def __init_subclass__(cls, **kwargs):
		super().__init_subclass__(**kwargs)
		if cls.Machine is not None:
			cls._known_machine_triggers = cls.Machine.process_machine_triggers(cls)


	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		if self.Machine is not None:
			self._known_gizmos = self.Machine.process_gizmos(self, self._known_machine_triggers)


	def create_depot(self, *args, **kwargs):
		return self.Depot(self._known_gizmos, *args, **kwargs)






























from typing import Iterator, Optional, Any, Iterable, Callable
from omnibelt import extract_function_signature
from omnibelt.crafts import AbstractSkill, NestableCraft

from .errors import GadgetFailure, MissingGadget
from .abstract import AbstractGadget, AbstractGaggle, AbstractGig


class GadgetBase(AbstractGadget):
	"""
	GadgetBase is a simple base class that adds two kinds of internal exceptions for gadgets to raise or catch as
	needed.

	Attributes:
		_GadgetFailure: The general exception that is raised when a gadget fails.
		_MissingGadgetError: The exception that is raised when a required gizmo is missing.
	"""
	_GadgetFailure = GadgetFailure
	_MissingGadgetError = MissingGadget


class SingleGadgetBase(GadgetBase):
	"""
	SingleGadgetBase is a simple bass class for gadgets that only produce a single gizmo, which is specified at init.

	Attributes:
		_gizmo (str): The gizmo that this gadget grabs.
	"""

	def __init__(self, gizmo: str, **kwargs):
		"""
		Initializes a new instance of the SingleGadgetBase class.

		Args:
			gizmo (str): The gizmo that this gadget produces.
			**kwargs: Arbitrary keyword arguments for superclasses.
		"""
		super().__init__(**kwargs)
		self._gizmo = gizmo

	def gizmos(self) -> Iterator[str]:
		"""
		Lists the gizmo that this gadget grabs.

		Returns:
			Iterator[str]: An iterator over the gizmo that this gadget grabs.
		"""
		yield self._gizmo


class FunctionGadget(SingleGadgetBase):
	"""
	FunctionGadget is a subclass of SingleGadgetBase for gadgets that produce a single gizmo using a given function.
	The function should take a single argument, the context (gig), and return the output gizmo.

	Attributes:
		_gizmo (str): The gizmo that this gadget grabs.
		_fn (Callable[[AbstractGig], Any]): The function that this gadget uses to grab the gizmo.
	"""

	def __init__(self, gizmo: str, fn: Callable[[AbstractGig], Any], **kwargs):
		"""
		Initializes a new instance of the FunctionGadget class.

		Args:
			gizmo (str): The gizmo that this gadget produces.
			fn (Callable[[AbstractGig], Any]): The function that this gadget uses to produce the gizmo.
			**kwargs: Arbitrary keyword arguments for superclasses.
		"""
		super().__init__(gizmo=gizmo, **kwargs)
		self._fn = fn

	def __repr__(self):
		"""
		Returns a string representation of this gadget. The representation includes the class name, the function name,
		and the gizmo.

		Returns:
			str: A string representation of this gadget.
		"""
		name = getattr(self._fn, '__qualname__', None)
		if name is None:
			name = getattr(self._fn, '__name__', None)
		return f'{self.__class__.__name__}({name}: {self._gizmo})'

	@property
	def __call__(self):
		"""
		Returns the function that this gadget uses to grab the gizmo.

		Returns:
			Callable[[AbstractGig], Any]: The function that this gadget uses to grab the gizmo.
		"""
		return self._fn

	def __get__(self, instance, owner):
		"""
		Returns the function that this gadget uses to grab the gizmo. This method is used to make the gadget callable.

		Args:
			instance: The instance that the function is called on.
			owner: The owner of the instance.

		Returns:
			Callable[[AbstractGig], Any]: The function that this gadget uses to grab the gizmo.
		"""
		return self._fn.__get__(instance, owner)

	def grab_from(self, ctx: Optional[AbstractGig], gizmo: str) -> Any:
		"""
		Returns the given gizmo from this gadget, or raises MissingGadgetError if the gizmo cannot be grabbed.

		Args:
			ctx (Optional[AbstractGig]): The context from which to grab the gizmo.
			gizmo (str): The gizmo to grab.

		Returns:
			Any: The grabbed gizmo.

		Raises:
			MissingGadgetError: If the wrong gizmo is requested.
		"""
		if gizmo != self._gizmo:
			raise self._MissingGadgetError(gizmo)
		return self._fn(ctx)


class AutoFunctionGadget(FunctionGadget):
	"""
	AutoFunctionGadget is a subclass of FunctionGadget that produces a single gizmo using a given function.
	The function can take any number of arguments, and any arguments that are gizmos will be grabbed from
	the gig and passed to the function automatically.
	The gizmo and the function are specified at initialization.

	Attributes:
		_gizmo (str): The gizmo that this gadget grabs.
		_fn (Callable[tuple, Any]): The function that this gadget uses to grab the gizmo.
	"""

	@staticmethod
	def _extract_gizmo_args(fn: Callable, ctx: AbstractGig, *, args: Optional[tuple] = None,
							kwargs: Optional[dict[str, Any]] = None) -> tuple[list[Any], dict[str, Any]]:
		"""
		Extracts the arguments for the function that this gadget uses to grab the gizmo. Any arguments that are gizmos
		are grabbed from the gig.

		Args:
			fn (Callable): The function that this gadget uses to produce the gizmo.
			ctx (AbstractGig): The context from which to grab any arguments needed by fn.
			args (Optional[tuple]): The positional arguments for the function passed in manually.
			kwargs (Optional[dict[str, Any]]): The keyword arguments for the function passed in manually.

		Returns:
			tuple[list[Any], dict[str, Any]]: A tuple containing a list of positional arguments and a dictionary
			of keyword arguments.
		"""
		return extract_function_signature(fn, args=args, kwargs=kwargs, default_fn=ctx.grab)

	def grab_from(self, ctx: Optional[AbstractGig], gizmo: str) -> Any:
		"""
		Returns the given gizmo from this gadget, or raises MissingGadgetError if the gizmo cannot be grabbed.
		The function that this gadget uses to grab the gizmo is called with the arguments extracted by _extract_gizmo_args.

		Args:
			ctx (Optional[AbstractGig]): The context from which to grab the gizmo.
			gizmo (str): The gizmo to grab.

		Returns:
			Any: The grabbed gizmo.

		Raises:
			MissingGadgetError: If the wrong gizmo is requested.
		"""
		if gizmo != self._gizmo:
			raise self._MissingGadgetError(gizmo)

		args, kwargs = self._extract_gizmo_args(self._fn, ctx)
		return self._fn(*args, **kwargs)





from typing import Tuple, Iterator, Dict

from .abstract import AbstractAssessible, AbstractAssessment, AbstractTool


class AbstractSignature:
	pass


class SimpleSignature(AbstractSignature):
	def __init__(self, output, inputs=(), meta=(), *, name=None, fn=None, **props):
		if isinstance(inputs, str):
			inputs = (inputs,)
		# inputs = tuple(inputs)
		if isinstance(meta, str):
			meta = (meta,)
		# meta = tuple(meta)
		if name is None and fn is not None:
			name = fn.__name__
		super().__init__()
		self.inputs = inputs
		self.meta = meta
		self.output = output
		self.props = props
		self.name = name
		self.fn = fn


	def __contains__(self, label: str):
		return label in self.inputs or label in self.meta or label == self.output


	def copy(self, output=None, inputs=None, meta=None, props=None, **kwargs):
		if output is None:
			output = self.output
		if inputs is None:
			inputs = self.inputs
		if meta is None:
			meta = self.meta
		if props is None:
			props = self.props
		return type(self)(output, inputs, meta, **{**props, **kwargs})


	def replace(self, fixes: Dict[str,str], **kwargs):
		new = self.copy(output=fixes.get(self.output, self.output), inputs=tuple(fixes.get(i, i) for i in self.inputs),
		                meta=tuple(fixes.get(m, m) for m in self.meta), **kwargs)
		return new


	def __str__(self):
		inp = ', '.join(self.inputs)
		ant = '()'
		if self.inputs and self.meta:
			raise NotImplementedError
			ant = f'{inp} ({", ".join(self.meta)})'
		elif self.meta:
			ant = ', '.join(f'<{m}>' for m in self.meta)
		elif self.inputs:
			ant = inp
		# msg = f'{self.output} <- {ant}'
		msg = f'{ant} -> {self.output}'
		if self.name:
			msg = f'{self.name}: {msg}'
		return msg


	def __repr__(self):
		return f'{self.__class__.__name__}({self.inputs} -> {self.output})'



class Signatured:
	_Signature = SimpleSignature

	def signatures(self, owner = None) -> Iterator['AbstractSignature']:
		raise NotImplementedError



class SimpleAssessment(AbstractAssessment):
	class Node:
		def __init__(self, node, **props):
			self.node = node
			self.props = props

		def __eq__(self, other):
			return id(self.node) == id(other.node)

		def __hash__(self):
			return id(self.node)


	class Edge:
		def __init__(self, src, dest, **props):
			self.src = src
			self.dest = dest
			self.props = props

		def __eq__(self, other):
			return (id(self.src), id(self.dest)) == (id(other.src), id(other.dest))

		def __hash__(self):
			return hash((id(self.src), id(self.dest)))


	def __init__(self):
		self.nodes = set()
		self.edges = set()


	def add_edge(self, src, dest, **props):
		self.edges.add(self.Edge(src, dest, **props))
		self.add_node(src)
		self.add_node(dest)


	def add_node(self, node, **props):
		self.nodes.add(self.Node(node, **props))





















from typing import Optional, Any, Iterator, Hashable, Type
from omnibelt import unspecified_argument, filter_duplicates

from ..structure import spaces

from .errors import ToolFailedError, MissingGizmoError



class Gizmoed:
	def gizmos(self) -> Iterator[str]:
		raise NotImplementedError


	def has_gizmo(self, gizmo: str) -> bool:
		raise NotImplementedError


	def vendors(self, gizmo: str) -> Iterator['AbstractTool']:
		yield from self.tools()


	def tools(self) -> Iterator['AbstractTool']:
		raise NotImplementedError



class SingleVendor(Gizmoed):
	def vendor(self, gizmo: str, default: Any = unspecified_argument) -> 'AbstractTool':
		raise NotImplementedError


	def vendors(self, gizmo: str) -> Iterator['AbstractTool']:
		yield self.vendor(gizmo)



class AbstractTool(Gizmoed): # leaf/source

	_ToolFailedError = ToolFailedError
	_MissingGizmoError = MissingGizmoError

	def tools(self) -> Iterator['AbstractTool']:
		yield self


	# def validate_context(self, ctx: 'AbstractContext'):
	# 	'''makes sure the context is valid for this tool, and returns a new context if necessary'''
	# 	return ctx


	def get_from(self, ctx: Optional['AbstractContext'], gizmo: str):
		raise NotImplementedError


	def get(self, gizmo: str, default: Any = unspecified_argument):
		try:
			return self.get_from(None, gizmo)
		except ToolFailedError:
			if default is unspecified_argument:
				raise
			return default


	def __getitem__(self, gizmo: str):
		return self.get_from(None, gizmo)


	def __contains__(self, gizmo: str):
		return self.has_gizmo(gizmo)



class AbstractKit(AbstractTool): # branch/router
	def tools(self) -> Iterator['AbstractTool']: # must iterate over children, not self (prevents infinite recursion)
		raise NotImplementedError


	def gizmos(self) -> Iterator[str]:
		yield from filter_duplicates(*[tool.gizmos() for tool in self.tools()])


	def has_gizmo(self, gizmo: str) -> bool:
		return any(tool.has_gizmo(gizmo) for tool in self.vendors(gizmo))


	def get_from(self, ctx: Optional['AbstractContext'], gizmo: str):
		tries = 0
		for tool in self.vendors(gizmo):
			try:
				return tool.get_from(ctx, gizmo)
			except ToolFailedError:
				tries += 1
		raise self._ToolFailedError(f'No tool for {gizmo} in {self} (tried {tries} tool/s)')



class AbstractSourcedKit(AbstractKit):
	def sources(self) -> Iterator['AbstractTool']:
		raise NotImplementedError


	def gizmos(self) -> Iterator[str]:
		yield from filter_duplicates(*[source.gizmos() for source in self.sources()])


	def tools(self) -> Iterator['AbstractTool']:
		for source in self.sources():
			yield from source.tools()



class AbstractContext(AbstractKit):
	@property
	def context_id(self) -> Hashable:
		raise NotImplementedError


	def __hash__(self):
		return hash(self.context_id)


	def __eq__(self, other):
		if isinstance(other, AbstractContext):
			return self.context_id == other.context_id
		return NotImplemented


	def get(self, gizmo: str, default: Any = unspecified_argument):
		try:
			return self.get_from(self, gizmo)
		except ToolFailedError:
			if default is unspecified_argument:
				raise
			return default


	def __getitem__(self, gizmo: str):
		return self.get_from(self, gizmo)



class AbstractDynamicKit(AbstractKit):
	def include(self, *sources: AbstractTool) -> 'AbstractDynamicKit':
		raise NotImplementedError



class AbstractDynamicContext(AbstractContext, AbstractDynamicKit):
	def include(self, *sources: AbstractTool) -> 'AbstractDynamicContext':
		raise NotImplementedError



class AbstractResourceable(AbstractTool):
	def as_resource(self, mogul: 'AbstractMogul') -> 'AbstractResource':
		raise NotImplementedError



class AbstractResource:
	def validate_context(self, ctx: AbstractContext) -> AbstractContext:
		return ctx



class AbstractScopable(AbstractResource, AbstractTool):
	'''
	 A tool can specify a scope for the given context to use when accessing its gizmos.

	 For example, datasets can specify a scope for the context to use when accessing their data.

	 Can raise a `InvalidContextError` if the context is not valid for this tool
	 (e.g. if it doesn't have the required data, like a `size`)
	 '''
	def create_scope(self, ctx: AbstractContext) -> Optional['AbstractTool']:
		return self


	def validate_context(self, ctx: AbstractContext) -> AbstractContext:
		scope = self.as_scope(ctx)
		return ctx if scope is None else ctx.include(scope)



class AbstractScopeGenerator(AbstractResource, AbstractTool):
	'''
	 A tool can specify a scope for the given context to use when accessing its gizmos.

	 For example, datasets can specify a scope for the context to use when accessing their data.

	 Can raise a `InvalidContextError` if the context is not valid for this tool
	 (e.g. if it doesn't have the required data, like a `size`)
	 '''
	def as_scope(self, ctx: AbstractContext) -> Optional['AbstractScope']:
		pass


	def validate_context(self, ctx: AbstractContext) -> AbstractContext:
		scope = self.as_scope(ctx)
		return ctx if scope is None else scope



class AbstractMogul: # controls/manages/creates/generates contexts => trainers, experiments, etc.
	def resources(self) -> Iterator[AbstractResource]:
		raise NotImplementedError



class AbstractScope(AbstractContext):
	def gizmoto(self) -> Iterator[str]: # iterates over external gizmos (products)
		yield from self.gizmos()


	def gizmo_to(self, external: str) -> str: # global -> local
		return external


	def gizmo_from(self, internal: str) -> str: # local -> global
		return internal



class AbstractReproducableScope(AbstractScopable, AbstractScope):
	pass



##########################################################################################



class AbstractSpaced:
	def space_of(self, gizmo: str) -> spaces.Dim:
		raise ToolFailedError(gizmo)



class AbstractChangableSpace(AbstractSpaced): # TODO: build into `space` crafts
	def change_space_of(self, gizmo: str, space: spaces.Dim):
		raise NotImplementedError


	def _missing_spaces(self):
		# yield from (gizmo for gizmo in self.gizmos() if self.space_is_missing(gizmo))
		raise NotImplementedError


	def clear_space(self, gizmo: str) -> bool:
		raise NotImplementedError



class AbstractSchema(AbstractSpaced): # context for building
	pass


##########################################################################################



class AbstractAssessment:
	def add_node(self, node, **props):
		raise NotImplementedError


	def add_edge(self, src, dest, **props):
		raise NotImplementedError


	def expand(self, node: 'AbstractAssessible'):
		node.assess(self)


	def expand_static(self, node: Type['AbstractAssessible']):
		node.assess_static(self)



class AbstractAssessible:
	@classmethod
	def assess_static(cls, assessment: AbstractAssessment):
		assessment.add_node(cls)


	def assess(self, assessment: AbstractAssessment):
		assessment.add_node(self)



class Loggable:
	@staticmethod
	def log(ctx):
		raise NotImplementedError









from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable, Type, Set
import math

# moguls generate contexts

from ..features import Prepared, Seedable, gen_deterministic_seed

from .abstract import AbstractMogul, AbstractContext, AbstractSourcedKit, AbstractResource, \
	AbstractTool, AbstractDynamicKit, AbstractScopable, AbstractResourceable, AbstractDynamicContext



class IterableMogul(AbstractMogul):
	def __iter__(self):
		raise NotImplementedError



class IteratorMogul(IterableMogul):
	def __iter__(self):
		return self


	def __next__(self):
		raise NotImplementedError


	def current_context(self):
		raise NotImplementedError


	def done(self) -> bool:
		'''not done guarantees that __next__ will return a new context'''
		raise NotImplementedError



class LimitMogul(IterableMogul):
	def __len__(self):
		raise NotImplementedError



class SelectionMogul(IterableMogul):
	def __getitem__(self, item):
		raise NotImplementedError



class BuildingContextMogul(AbstractMogul):
	def build_context(self) -> AbstractContext:
		raise NotImplementedError



class ValidationMogul(AbstractMogul):
	def validate_context(self, ctx: AbstractContext) -> AbstractContext:
		return ctx



########################################################################################################################



# class SimpleMogul(AbstractMogul):
# 	def __init__(self, resources=None, **kwargs):
# 		if resources is None:
# 			resources = []
# 		super().__init__(**kwargs)
# 		self._resources = resources
#
#
# 	def resources(self) -> Iterator[AbstractResource]:
# 		yield from reversed(self._resources)




class DefaultResourceMogul(AbstractMogul, AbstractDynamicKit):
	class _DefaultResource(AbstractResource):
		def __init__(self, source: AbstractTool, **kwargs):
			super().__init__(**kwargs)
			self._source = source


		@property
		def source(self) -> AbstractTool:
			return self._source


		def validate_context(self, ctx: AbstractDynamicContext) -> AbstractContext:
			return ctx.include(self.source)


	def __init__(self, *, resources=None, **kwargs):
		if resources is None:
			resources = []
		super().__init__(**kwargs)
		self._resources = resources


	def resources(self) -> Iterator[AbstractResource]:
		yield from reversed(self._resources)


	def _as_resource(self, source: AbstractTool):
		if isinstance(source, AbstractResource):
			return source
		resource = source.as_resource(self) if isinstance(source, AbstractResourceable) else None
		if resource is None:
			resource = self._DefaultResource(source)
		return resource


	def include(self, *sources: AbstractTool):
		for source in sources:
			self._resources.append(self._as_resource(source))
		return self



class CreativeMogul(AbstractMogul):
	_Context = None


	def _create_context(self, *args, **kwargs):
		return self._Context(*args, **kwargs)



class ValidatedCreativeMogul(CreativeMogul, AbstractResource):
	def _validate_context(self, ctx: AbstractContext) -> AbstractContext: # as a driver or an observer
		for resource in self.resources():
			ctx = resource.validate_context(ctx)
		return ctx


	def _create_context(self, *args, **kwargs): # as a driver
		return self._validate_context(super()._create_context(*args, **kwargs))


	def validate_context(self, ctx: AbstractContext) -> AbstractContext: # as an observer
		return self._validate_context(ctx)



class SeedingMogul(DefaultResourceMogul, Seedable):
	class _Context_Seeder(AbstractResource, Seedable):
		def __init__(self, seed: int, make_default: Optional[bool] = True, **kwargs):
			super().__init__(**kwargs)
			self._start_seed = seed
			self._steps = 0
			self._current_seed = seed
			self._make_default = make_default


		@property
		def seed(self) -> int:
			return self._start_seed


		def next_seed(self) -> int:
			self._current_seed = gen_deterministic_seed(self._current_seed)
			self._steps += 1
			return self._current_seed


		def validate_context(self, ctx: Seedable) -> AbstractContext:
			ctx.reset_rng(self.next_seed())
			return ctx


	def __init__(self, context_seeder=None, **kwargs):
		super().__init__(**kwargs)
		if context_seeder is None:
			context_seeder = self._Context_Seeder(self.seed)
		self._context_seeder = context_seeder
		self.include(self._context_seeder)



class SimpleMogul(DefaultResourceMogul, ValidatedCreativeMogul):
	pass



########################################################################################################################



class OptimMogul(IterableMogul):
	@property
	def iteration_count(self) -> int: # of the optimizer
		raise NotImplementedError



class OptimBudgetMogul(OptimMogul):
	@property
	def budget_iterations(self) -> Optional[int]:
		raise NotImplementedError


	@property
	def remaining_iterations(self):
		return self.budget_iterations - self.iteration_count



class BatchMogul(IterableMogul):
	@property
	def batch_size(self):
		raise NotImplementedError


	@property
	def sample_count(self) -> int:
		raise NotImplementedError
	@property
	def batch_count(self) -> int:
		raise NotImplementedError



class EpochStatMogul(BatchMogul, LimitMogul):
	@property
	def epoch_size(self):
		raise NotImplementedError


	@property
	def current_epoch(self) -> int:
		raise NotImplementedError


	@property
	def completed_epochs(self) -> int:
		raise NotImplementedError



# class BudgetMogul(BatchBudgetMogul, OptimBudgetMogul):
# 	pass



########################################################################################################################



# class SimpleTrainer(DynamicMogul):
# 	def __init__(self, model, *, optim=None, **kwargs):
# 		if optim is None:
# 			optim = model.default_optim()
# 		super().__init__(**kwargs)
# 		self.model = model
# 		self.optim = optim
#
#
# 	def fit(self, dataset):
# 		for ctx in self.iterate(dataset):
# 			out = self.step(ctx)
# 		raise NotImplementedError
# 		# return out
#
#
# 	@staticmethod
# 	def iterate(dataset):
# 		yield from dataset
#
#
# 	def step(self, ctx):
# 		self.optim.step(ctx)
#
#
#
# class Checkpoint(SimpleTrainer):
# 	def checkpoint(self, ctx):
# 		pass
#
#
#
# class Evaluatable(SimpleTrainer):
# 	def evaluate(self, dataset): # valset or testset
# 		pass











try:
	from graphviz import Digraph

except ImportError:
	pass




def signature_graph(kit, name='unknown'):

	g = Digraph(name, filename=f'{name}.gv')

	for sig in kit.signatures():
		for inp in sig.inputs:
			g.edge(inp, sig.output)
		for meta in sig.meta:
			g.edge(meta, sig.output, style='dashed')
			g.node(meta, shape='box')

	return g

















from .abstract import AbstractGadget, AbstractGaggle, AbstractGig, AbstractGang
from .errors import GadgetFailure, MissingGadget
from .tools import ToolCraft, AutoToolCraft, ToolDecorator, AutoToolDecorator
from .gizmos import DashGizmo
from .gaggles import MutableGaggle, LoopyGaggle, CraftyGaggle
from .gigs import CacheGig, GroupCache
from .groups import GroupBase, CachableGroup, SelectiveGroup



class tool(AutoToolDecorator):
	"""
	The tool class is a subclass of AutoToolDecorator. It provides a convenient way to create tool instances
	with automatic function gadget capabilities. It does not specify a gizmo type, meaning it can handle any type of gizmo.

	Attributes:
		_gizmo_type (None): The type of the gizmo. Defaults to None.
	"""

	# _gizmo_type = DashGizmo

	class from_context(ToolDecorator):
		"""
		The from_context class is a nested class within tool that inherits from ToolDecorator. It provides a way to create
		tool instances that handle DashGizmo types specifically. This is useful when the gizmos being handled contain dashes.

		Attributes:
			_gizmo_type (DashGizmo): The type of the gizmo. Defaults to DashGizmo.
		"""
		# _gizmo_type = DashGizmo
		pass


class ToolKit(LoopyGaggle, MutableGaggle, CraftyGaggle):
	"""
	The ToolKit class is a subclass of LoopyGaggle, MutableGaggle, and CraftyGaggle. It provides methods to handle
	tools in a kit.

	Methods:
		__init__(self, *args, **kwargs): Initializes a new instance of the ToolKit class.
	"""

	def __init__(self, *args, **kwargs):
		"""
		Initializes a new instance of the ToolKit class.

		This method initializes the superclass with the provided arguments and processes the crafts in the tool kit.

		Args:
			args: Variable length argument list.
			kwargs: Arbitrary keyword arguments.
		"""
		super().__init__(*args, **kwargs)
		self._process_crafts()


class Context(GroupCache, LoopyGaggle, MutableGaggle, AbstractGig):
	"""
	The Context class is a subclass of GroupCache, LoopyGaggle, MutableGaggle, and AbstractGig. It provides methods to handle
	gadgets in a context.

	Methods:
		__init__(self, *gadgets: AbstractGadget, **kwargs): Initializes a new instance of the Context class.
		__getitem__(self, item): Returns the grabbed item from the context.
	"""

	# _gizmo_type = DashGizmo # TODO

	def __init__(self, *gadgets: AbstractGadget, **kwargs):
		"""
		Initializes a new instance of the Context class.

		This method initializes the superclass with the provided arguments and includes the provided gadgets in the context.

		Args:
			gadgets (AbstractGadget): The gadgets to be included in the context.
			kwargs: Arbitrary keyword arguments.
		"""
		super().__init__(**kwargs)
		self.include(*gadgets)

	def __getitem__(self, item):
		"""
		Returns the grabbed item from the context.

		Args:
			item: The item to be grabbed from the context.

		Returns:
			Any: The grabbed item from the context.
		"""
		return self.grab(item)


class Scope(CachableGroup, LoopyGaggle, MutableGaggle, AbstractGang):
	"""
	The Scope class is a subclass of CachableGroup, LoopyGaggle, MutableGaggle, and AbstractGang. It provides methods
	to handle gadgets in a scope.

	Methods:
		__init__(self, *gadgets: AbstractGadget, **kwargs): Initializes a new instance of the Scope class.
	"""

	def __init__(self, *gadgets: AbstractGadget, **kwargs):
		"""
		Initializes a new instance of the Scope class.

		This method initializes the superclass with the provided arguments and includes the provided gadgets in the scope.

		Args:
			gadgets (AbstractGadget): The gadgets to be included in the scope.
			kwargs: Arbitrary keyword arguments.
		"""
		super().__init__(**kwargs)
		self.include(*gadgets)

	def __getitem__(self, item):
		"""
		Returns the grabbed item from the context.

		Args:
			item: The item to be grabbed from the context.

		Returns:
			Any: The grabbed item from the context.
		"""
		return self.grab(item)

class Selection(SelectiveGroup, Scope):
	"""
	The Selection class is a subclass of SelectiveGroup and Scope. It provides methods to handle selective gizmo mapping
	within a scope.

	Methods:
		__init__(self, *gadgets: AbstractGadget, **kwargs): Initializes a new instance of the Selection class.
	"""

	def __init__(self, *gadgets: AbstractGadget, **kwargs):
		"""
		Initializes a new instance of the Selection class.

		This method initializes the superclass with the provided arguments and includes the provided gadgets in the selection.

		Args:
			gadgets (AbstractGadget): The gadgets to be included in the selection.
			kwargs: Arbitrary keyword arguments.
		"""
		super().__init__(*gadgets, **kwargs)

	def __getitem__(self, item):
		"""
		Returns the grabbed item from the context.

		Args:
			item: The item to be grabbed from the context.

		Returns:
			Any: The grabbed item from the context.
		"""
		return self.grab(item)


from .imports import *
from .errors import *


class AbstractBlueprint:
	pass



class AbstractArtist:
	def details(self) -> Dict[str, Any]:
		raise NotImplementedError


	def derivative(self, *args, **kwargs) -> 'AbstractArtist':
		'''Returns a new artist that is a derivative of this one (ie. same except for the given args, and kwargs).'''
		raise NotImplementedError


	@agnostic
	def draft(self, *args, **kwargs) -> Type: # aka "product"
		if isinstance(self, type):
			return self(*args, **kwargs).design()
		return self._design()


	@agnostic
	def describe(self, *args, **kwargs): # aka "plan"
		pass


	@agnostic
	def develop(self, *args, **kwargs): # aka "build"
		pass


	@agnostic
	def design(self, *args, **kwargs): # aka "draft"
		pass


	def _design(self):
		pass


	def _develop(self):
		raise NotImplementedError



class Artist(AbstractArtist, Replicator):
	def derivative(self, *args, **kwargs) -> 'AbstractArtist':
		'''Returns a new artist that is a derivative of this one (ie. same except for the given args, and kwargs).'''
		kwargs = args2kwargs(self.__class__, args, kwargs)
		return self.replicate(**kwargs)



class AbstractArchitect(AbstractArtist):
	def design_with(self, blueprint: Any):
		raise NotImplementedError


	def design_for(self, owner: Any) -> Any:
		raise NotImplementedError






























# import random
from omnibelt import agnosticmethod
from . import operations


class Measure(operations.Metric):
	@staticmethod
	def difference(x, y): # should find an z such that z + y = x
		return x - y


	@agnosticmethod
	def measure(self, x, y):
		return self.difference(x, y).abs()


	@staticmethod
	def distance(x, y):
		raise NotImplementedError



class Norm(Measure):
	@staticmethod
	def magnitude(x):
		raise NotImplementedError


	@agnosticmethod
	def distance(self, x, y):
		return self.magnitude(self.difference(x, y))



class Lp(Norm):
	p = None

	@agnosticmethod
	def magnitude(self, x, dim=1):
		return x.norm(dim=dim, p=self.p)


	def __str__(self):
		return f'L{self.p}()'


class L0(Lp): p = 0
class L1(Lp): p = 1
class L2(Lp): p = 2
class Linf(Lp): p = float('inf')







from omnibelt import agnostic

from ..features import Prepared
from ..parameters.abstract import AbstractBuilder



class AbstractEvaluatable:
	@staticmethod
	def evaluate(source, **kwargs):
		raise NotImplementedError



class AbstractFitable:
	@staticmethod
	def fit(source, **kwargs):
		raise NotImplementedError



class AbstractTrainable(AbstractFitable, Prepared):
	_Trainer = None
	def fit(self, source, **kwargs):
		self.prepare(source)
		trainer = self._Trainer(self)
		return trainer.fit(source=source, **kwargs)


	def step(self, batch):
		pass



class AbstractModel(AbstractFitable, AbstractEvaluatable):
	pass



class AbstractTrainableModel(AbstractModel, AbstractTrainable):
	pass



class AbstractTrainer(AbstractFitable, AbstractEvaluatable):
	def __init__(self, model=None, **kwargs):
		super().__init__(**kwargs)
		if model is not None:
			self.set_model(model)


	def set_model(self, model):
		raise NotImplementedError


	def loop(self, source, **kwargs):
		raise NotImplementedError


	def step(self, batch):
		raise NotImplementedError








from .imports import *
from .errors import *

T = TypeVar('T')



class AbstractQuirk:
	def setup(self, owner: Type[T], name: str) -> 'AbstractQuirk': # TODO: Self
		return self


	def resolve(self, instance: Optional[T] = None, owner: Optional[Type[T]] = None) -> Any:
		raise NotImplementedError


	def reset(self, instance: T, value: Optional[Any] = unspecified_argument) -> Any:
		raise NotImplementedError


	def rip(self, instance: Optional[T] = None) -> Any:
		raise NotImplementedError


	def replicate(self, **kwargs) -> Any:
		return self.__class__(**kwargs)



class ReplicatorQuirk(AbstractQuirk, Replicator):
	pass



# class ReferenceQuirk(AbstractQuirk):
# 	def reference(self, **update):
# 		raise NotImplementedError



class DescriptorQuirk(ReplicatorQuirk):
	def __init__(self, *, attrname: Optional[str] = None, **kwargs):
		super().__init__(**kwargs)
		self._attrname = attrname


	_MissingValueError = MissingValueError
	_MissingOwnerError = MissingOwnerError
	def setup(self, owner: Type[T], name: str) -> 'AbstractQuirk':
		if self._attrname is None:
			self._attrname = name
		elif name != self._attrname:
			raise TypeError(
				f"Cannot assign the same {self.__class__.__qualname__} to two different names "
				f"({self._attrname!r} and {name!r})."
			)
		return self


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'attrname' not in kwargs:
			kwargs['attrname'] = self._attrname
		return kwargs


	def __str__(self):
		return f'<quirk:{self._attrname}>'


	def __repr__(self):
		return f'{self.__class__.__qualname__}({self._attrname})'


	def __set_name__(self, owner: Type[T], name: str):
		self.setup(owner, name)


	def __get__(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		return self.resolve(instance, owner=owner)


	def __set__(self, instance: T, value: Any):
		self.reset(instance, value)


	def __delete__(self, instance: T):
		self.rip(instance)



class OwnedDescriptorQuirk(DescriptorQuirk):
	def __init__(self, owner: Optional[Type[T]] = None, **kwargs):
		super().__init__(**kwargs)
		self._owner = owner


	def setup(self, owner: Type[T], name: str) -> 'AbstractQuirk': # TODO: Self
		self._owner = owner
		return super().setup(owner, name)


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'owner' not in kwargs:
			kwargs['owner'] = self._owner
		return super()._replicator_kwargs(kwargs)


	def __str__(self):
		return f'<quirk:{self._owner.__qualname__}.{self._attrname}>'


	def __repr__(self):
		return f'{self.__class__.__name__}({self._owner.__qualname__}.{self._attrname})'



class DefaultQuirk(DescriptorQuirk):
	_missing_value = object()
	def __init__(self, default: Optional[Any] = unspecified_argument, *,
	             getter: Optional[Callable[[T], Any]] = None, **kwargs):
		if default is unspecified_argument:
			default = self._missing_value
		getter, default = self._check_fget(getter, default)
		super().__init__(**kwargs)
		self._default = default
		self._getter = getter


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'default' not in kwargs:
			kwargs['default'] = self._default
		if 'getter' not in kwargs:
			kwargs['getter'] = self._getter
		return super()._replicator_kwargs(kwargs)


	def _check_fget(self, getter: Optional[Callable[[T], Any]] = None, default: Optional[Any] = unspecified_argument) \
			-> Tuple[Optional[Callable[[T], Any]], Optional[Any]]:
		if getter is None:
			if callable(default) and not isinstance(default, type) and default.__qualname__ != default.__name__:
				return default, self._missing_value  # decorator has no other args
			return None, default  # no getter provided (optionally can be added with __call__)
		return getter, default  # getter was specified as keyword argument


	def realize(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		if self._default is self._missing_value:
			raise self._MissingValueError(self)
		return self._default


	def resolve(self, instance: Optional[T] = None, owner: Optional[Type[T]] = None) -> Any:
		if self._getter is None:
			return self.realize(instance, owner=owner)
		if instance is None:
			raise self._MissingOwnerError(self)
		return self._getter.__get__(instance, owner=owner)()


	def getter(self, getter: Callable[[T], Any]) -> 'DefaultQuirk':
		return self.replicate(getter=getter)



class SetterQuirk(DescriptorQuirk):
	def __init__(self, *, setter: Optional[Callable[[T, Any], Any]] = None, **kwargs):
		super().__init__(**kwargs)
		self._setter = setter


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'setter' not in kwargs:
			kwargs['setter'] = self._setter
		return super()._replicator_kwargs(kwargs)


	def reset(self, instance: T, value: Optional[Any] = unspecified_argument) -> Any:
		if instance is None:
			raise self._MissingOwnerError(self)
		if value is unspecified_argument:
			raise self._MissingValueError(self)
		if self._setter is not None:
			return self._setter.__get__(instance)(value)
		return super().reset(instance, value)


	def setter(self, setter: Callable[[T, Any], Any]) -> 'SetterQuirk':
		return self.replicate(setter=setter)



class DeleterQuirk(DescriptorQuirk):
	def __init__(self, *, deleter: Optional[Callable[[T], Any]] = None, **kwargs):
		super().__init__(**kwargs)
		self._deleter = deleter


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'deleter' not in kwargs:
			kwargs['deleter'] = self._deleter
		return super()._replicator_kwargs(kwargs)


	def rip(self, instance: Optional[T] = None) -> Any:
		if instance is None:
			raise MissingOwnerError(self)
		if self._deleter is not None:
			return self._deleter.__get__(instance)()
		return super().rip(instance)


	def deleter(self, deleter: Callable[[T], Any]) -> 'DeleterQuirk':
		return self.replicate(deleter=deleter)



class CachedQuirk(DescriptorQuirk):
	def __init__(self, cache: bool = True, **kwargs):
		super().__init__(**kwargs)
		self._cache = cache


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'cache' not in kwargs:
			kwargs['cache'] = self._cache
		return super()._replicator_kwargs(kwargs)


	def is_cached(self, instance: T) -> bool:
		return self._attrname in instance.__dict__


	def clear_cache(self, instance: T) -> 'CachedQuirk': # TODO: change to Self
		if self.is_cached(instance):
			del instance.__dict__[self._attrname]
		return self


	def _resolve(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		return super().resolve(instance, owner=owner)


	def _cache_value(self, instance: T, value: Any) -> Any:
		if instance is None:
			raise TypeError(f'Cannot cache {self} on class.')
		instance.__dict__[self._attrname] = value
		return value


	def resolve(self, instance: Optional[T] = None, owner: Optional[Type[T]] = None) -> Any:
		if self.is_cached(instance):
			# NOTE: this is usually redundant, since descriptor is called after __dict__ is checked
			return instance.__dict__[self._attrname]

		value = self._resolve(instance, owner=owner)
		return self._cache_value(instance, value) if self._cache else value


	def reset(self, instance: T, value: Optional[Any] = unspecified_argument) -> Any:
		self.clear_cache(instance)
		if value is not unspecified_argument:
			if self._cache:
				instance.__dict__[self._attrname] = value
				return value
			return super().reset(instance, value)


	def rip(self, instance: Optional[T] = None) -> Any:
		if self.is_cached(instance):
			del instance.__dict__[self._attrname]
		return super().rip(instance)



class UpdateQuirk(AbstractQuirk):
	def __init__(self, *, updater: Optional[Callable[[T, Any], Any]] = None, **kwargs):
		super().__init__(**kwargs)
		self._updater = updater


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'updater' not in kwargs:
			kwargs['updater'] = self._updater
		return super()._replicator_kwargs(kwargs)


	def revise(self, instance: T, value: Any) -> Any:
		if self._updater is None:
			return value
		return self._updater.__get__(instance)(value)


	_IgnoreResetFlag = IgnoreResetFlag
	def reset(self, instance: T, value: Optional[Any] = unspecified_argument) -> Any:
		try:
			if value is not unspecified_argument:
				value = self.revise(instance, value)
		except self._IgnoreResetFlag:
			if value is not unspecified_argument:
				return value
		else:
			return super().reset(instance, value)


	def update(self, updater: Callable[[T, Any], Any]) -> 'UpdateQuirk':
		return self.replicate(updater=updater)



class InheritableQuirk(AbstractQuirk):
	def inherit(self, owner: Union[T, Type[T]]) -> bool:
		return False



class HiddenQuirk(AbstractQuirk):
	def hidden(self, owner: Union[T, Type[T]]) -> bool:
		return False



class SimpleQuirk(UpdateQuirk, CachedQuirk, DefaultQuirk, SetterQuirk, DeleterQuirk, InheritableQuirk, HiddenQuirk):
	def __init__(self, *, inherit: bool = False, hidden: bool = False, **kwargs):
		super().__init__(**kwargs)
		self._inherit = inherit
		self._hidden = hidden


	# def resolve(self, instance: Optional[T] = None, owner: Optional[Type[T]] = None) -> Any:
	# 	if instance is None:
	# 		return self
	# 	return super().resolve(instance, owner=owner)


	def _replicator_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
		if 'inherit' not in kwargs:
			kwargs['inherit'] = self._inherit
		if 'hidden' not in kwargs:
			kwargs['hidden'] = self._hidden
		return super()._replicator_kwargs(kwargs)


	def inherit(self, owner: Union[T, Type[T]]) -> bool:
		return self._inherit


	def hidden(self, owner: Union[T, Type[T]]) -> bool:
		return self._hidden












